"""
Global constants shared by different modules.

.. versionadded: 23.1.1
"""

from typing import Final


#: Representation for redacted secrets
SECRET_REPR: Final[str] = "*******"

#: Key used in the field metadata
METADATA_KEY: Final[str] = "typed-settings"

#: Key for argparse option within Typed Settings' metadata
ARGPARSE_METADATA_KEY: Final[str] = "argparse"

#: Key for click option within Typed Settings' metadata
CLICK_METADATA_KEY: Final[str] = "click"



"""
Tests for "typed_settings.cls_utils".
"""

import dataclasses
from typing import Callable, List, Optional

import attrs
import pydantic
import pytest

from typed_settings import _compat, cls_utils, types


@attrs.define
class AttrsCls:
    """
    Test class for "attrs".
    """

    x: int = 0


@dataclasses.dataclass
class DataclassCls:
    """
    Test class for "dataclass".
    """

    x: int


class PydanticCls(pydantic.BaseModel):
    """
    Test class for "pydantic".
    """

    x: int


class NormalClass:
    """
    Test class.
    """

    x: int


@pytest.mark.parametrize("cls", [AttrsCls])
def test_deep_options(cls: type) -> None:
    """
    "deep_options()" converts similar classes of all suported class libs to the same
    result.
    """
    option_list = cls_utils.deep_options(cls)
    assert option_list == (
        types.OptionInfo(
            parent_cls=cls,
            path="x",
            cls=int,
            default=0,
            has_no_default=False,
            default_is_factory=False,
        ),
    )


def test_deep_options_typerror() -> None:
    """
    A TypeError is raised for non supported classes.
    """

    class C:
        x: int = 0

    pytest.raises(TypeError, cls_utils.deep_options, C)


@pytest.mark.parametrize("kind", ["attrs", "dataclasses", "pydantic"])
def test_resolve_types_decorator(kind: str) -> None:
    """
    The "resolve_type" function can be used as class decorator for all supported
    class libs.
    """
    if kind == "attrs":

        @cls_utils.resolve_types
        @attrs.define
        class NestedA:
            x: "int"

        @cls_utils.resolve_types(globalns=globals(), localns=locals())
        @attrs.define
        class A:
            opt: "List[NestedA]"

        assert attrs.fields(NestedA).x.type is int
        assert attrs.fields(A).opt.type is List[NestedA]

    elif kind == "dataclasses":

        @cls_utils.resolve_types
        @dataclasses.dataclass
        class NestedB:
            x: "int"

        @cls_utils.resolve_types(globalns=globals(), localns=locals())
        @dataclasses.dataclass
        class B:
            opt: "List[NestedB]"

        assert dataclasses.fields(NestedB)[0].type is int
        assert dataclasses.fields(B)[0].type is List[NestedB]

    elif kind == "pydantic":

        @cls_utils.resolve_types
        class NestedC(pydantic.BaseModel):
            x: "int"

        @cls_utils.resolve_types(globalns=globals(), localns=locals())
        class C(pydantic.BaseModel):
            opt: "List[NestedC]"

        assert NestedC.model_fields["x"].annotation is int
        assert C.model_fields["opt"].annotation is List[NestedC]

    else:
        pytest.fail(f"Invalid kind: {kind}")


@pytest.mark.parametrize(
    "cls, expected",
    [
        (AttrsCls, True),
        (DataclassCls, True),
        (PydanticCls, True),
        (NormalClass, False),
        (int, False),
        (float, False),
        (str, False),
        (list, False),
        (dict, False),
    ],
)
def test_handler_exists(cls: type, expected: bool) -> None:
    """
    "handler_exists()" return "True" for classes of a supported
    lib (attrs, dataclasses, Pydantic), but "False" for everything else.
    """
    assert cls_utils.handler_exists(cls) is expected


class TestGroupOptions:
    """
    Tests for "group_options()".
    """

    def test_typerror(self) -> None:
        """
        A TypeError is raised for non supported classes.
        """

        class C:
            x: int = 0

        pytest.raises(TypeError, cls_utils.group_options, [])

    def test_only_scalars(self) -> None:
        """
        If there are only scalar settings, create s single group.
        """

        @attrs.define
        class Parent:
            a: str
            b: int

        opts = cls_utils.deep_options(Parent)
        grouped = cls_utils.group_options(Parent, opts)
        assert grouped == [
            (Parent, opts[0:2]),
        ]

    def test_nested(self) -> None:
        """
        Create one group for the parent class' attributs and one for each
        nested class.
        """

        @attrs.define
        class Child:
            x: float
            y: int

        @attrs.define
        class Child2:
            x: str
            y: str

        @attrs.define
        class Parent:
            a: int
            b: float
            c: Child
            d: Child2

        opts = cls_utils.deep_options(Parent)
        grouped = cls_utils.group_options(Parent, opts)
        assert grouped == [
            (Parent, opts[0:2]),
            (Child, opts[2:4]),
            (Child2, opts[4:6]),
        ]

    def test_mixed(self) -> None:
        """
        If the parent class' attributes are not orderd, multiple groups for
        the main class are genererated.
        """

        @attrs.define
        class Child:
            x: float

        @attrs.define
        class Child2:
            x: str

        @attrs.define
        class Parent:
            a: int
            c: Child
            b: float
            d: Child2

        opts = cls_utils.deep_options(Parent)
        grouped = cls_utils.group_options(Parent, opts)
        assert grouped == [
            (Parent, opts[0:1]),
            (Child, opts[1:2]),
            (Parent, opts[2:3]),
            (Child2, opts[3:4]),
        ]

    def test_duplicate_nested_cls(self) -> None:
        """
        If the same nested class appears multiple times (in direct succession),
        create *different* groups for each attribute.
        """

        @attrs.define
        class Child:
            x: float
            y: int

        @attrs.define
        class Parent:
            b: Child
            c: Child

        opts = cls_utils.deep_options(Parent)
        grouped = cls_utils.group_options(Parent, opts)
        assert grouped == [
            (Child, opts[0:2]),
            (Child, opts[2:4]),
        ]

    def test_deep_nesting(self) -> None:
        """
        Grouping options only takes top level nested classes into account.
        """

        @attrs.define
        class GrandChild:
            x: int

        @attrs.define
        class Child:
            x: float
            y: GrandChild

        @attrs.define
        class Child2:
            x: GrandChild
            y: GrandChild

        @attrs.define
        class Parent:
            c: Child
            d: Child2

        opts = cls_utils.deep_options(Parent)
        grouped = cls_utils.group_options(Parent, opts)
        assert grouped == [
            (Child, opts[0:2]),
            (Child2, opts[2:4]),
        ]


class TestAttrs:
    """Tests for attrs classes."""

    def test_check_true(self) -> None:
        """
        "check()" detects "attrs" classes.
        """

        @attrs.define
        class C:
            x: int

        assert cls_utils.Attrs.check(C)

    def test_check_false(self) -> None:
        """
        "check()" only detects "attrs" classes.
        """

        class C:
            x: int

        assert not cls_utils.Attrs.check(C)

    def test_check_not_installed(self, unimport: Callable[[str], None]) -> None:
        """
        "check()" returns ``False`` if attrs is not installed.
        """

        @attrs.define
        class C:
            x: int

        unimport("attrs")

        assert not cls_utils.Attrs.check(C)

    @pytest.mark.skipif(not _compat.PY_39, reason="Doesn't work on 3.8")
    def test_iter_fields(self) -> None:
        """
        "iter_fields()" yields an option info for all options, including nested
        classes.
        """

        def factory_fn() -> Optional[int]:
            return None  # pragma: no cover

        @attrs.define
        class GrandChild:
            x: Optional[int] = attrs.field(
                factory=factory_fn,
                metadata={
                    "typed-settings": {
                        "help": "grand child x",
                        "argparse": {"metavar": "GX"},
                        "click": {"metavar": "GX"},
                    },
                },
            )

        @attrs.define
        class Child:
            x: "float" = attrs.field(  # Test resolving types
                metadata={
                    "typed-settings": {
                        "click": {"help": "child x", "metavar": "X"},
                    },
                },
            )
            y: GrandChild

        @attrs.define
        class Parent:
            x: str
            y: Child
            z: list[str] = ["default"]

        option_infos = cls_utils.Attrs.iter_fields(Parent)
        assert option_infos == (
            types.OptionInfo(
                parent_cls=Parent,
                path="x",
                cls=str,
                default=attrs.NOTHING,
                has_no_default=True,
                default_is_factory=False,
            ),
            types.OptionInfo(
                parent_cls=Child,
                path="y.x",
                cls=float,
                default=attrs.NOTHING,
                has_no_default=True,
                default_is_factory=False,
                metadata={
                    "click": {"help": "child x", "metavar": "X"},
                },
            ),
            types.OptionInfo(
                parent_cls=GrandChild,
                path="y.y.x",
                cls=Optional[int],  # type: ignore[arg-type]
                default=attrs.Factory(factory_fn),
                has_no_default=False,
                default_is_factory=True,
                metadata={
                    "argparse": {"help": "grand child x", "metavar": "GX"},
                    "click": {"help": "grand child x", "metavar": "GX"},
                    "help": "grand child x",
                },
            ),
            types.OptionInfo(
                parent_cls=Parent,
                path="z",
                cls=list[str],
                default=["default"],
                has_no_default=False,
                default_is_factory=False,
            ),
        )

    def test_unresolved_types(self) -> None:
        """Raise a NameError when types cannot be resolved."""

        @attrs.define
        class C:
            name: str
            x: "X"  # type: ignore  # noqa: F821

        with pytest.raises(NameError, match="name 'X' is not defined"):
            cls_utils.Attrs.iter_fields(C)

    def test_direct_recursion(self) -> None:
        """
        We do not (and cannot easily) detect recursion.  A NameError is already
        raised when we try to resolve all types.  This is good enough.
        """

        @attrs.define
        class Node:
            name: str
            child: "Node"

        with pytest.raises(NameError, match="name 'Node' is not defined"):
            cls_utils.Attrs.iter_fields(Node)

    def test_indirect_recursion(self) -> None:
        """
        We cannot (easily) detect indirect recursion but it is an error
        nonetheless.  This is not Dark!
        """

        @attrs.define
        class Child:
            name: str
            parent: "Parent"

        @attrs.define
        class Parent:
            name: str
            child: "Child"

        with pytest.raises(NameError, match="name 'Child' is not defined"):
            cls_utils.Attrs.iter_fields(Parent)

    def test_no_init_no_option(self) -> None:
        """
        No option is generated for an attribute if "init=False".
        """

        @attrs.frozen
        class Nested1:
            a: int = 0
            nb1: int = attrs.field(init=False)

        @attrs.frozen
        class Nested2:
            a: int = 0
            nb2: int = attrs.field(init=False)

        @attrs.frozen
        class Settings:
            a: int = 0
            na: int = attrs.field(init=False)
            n1: Nested1 = Nested1()
            n2: Nested2 = Nested2()

        options = [o.path for o in cls_utils.Attrs.iter_fields(Settings)]
        assert options == ["a", "n1.a", "n2.a"]

    def test_fields_to_parent_classes(self) -> None:
        """
        If there are only scalar settings, create s single group.
        """

        @attrs.define
        class Child1:
            x: int

        @attrs.define
        class Child2:
            x: float

        @attrs.define
        class Parent:
            a: str
            b: Child1
            c: Child2
            d: int

        result = cls_utils.Attrs.fields_to_parent_classes(Parent)
        assert result == {
            "a": Parent,
            "b": Child1,
            "c": Child2,
            "d": Parent,
        }


class TestDataclasses:
    """Tests for dataclasses."""

    def test_check_true(self) -> None:
        """
        "check()" detects dataclasses.
        """

        @dataclasses.dataclass
        class C:
            x: int

        assert cls_utils.Dataclasses.check(C)

    def test_check_false(self) -> None:
        """
        "check()" only detects dataclasses.
        """

        class C:
            x: int

        assert not cls_utils.Dataclasses.check(C)

    def test_iter_fields(self) -> None:
        """
        "iter_fields()" yields an option info for all options, including nested
        classes.
        """

        @dataclasses.dataclass
        class GrandChild:
            x: Optional[int] = dataclasses.field(
                default=None,
                metadata={
                    "typed-settings": {
                        "help": "grand child x",
                        "argparse": {"metavar": "GX"},
                        "click": {"metavar": "GX"},
                    },
                },
            )

        @dataclasses.dataclass
        class Child:
            x: "float" = dataclasses.field(  # Test resolving types
                metadata={
                    "typed-settings": {
                        "click": {"help": "child x", "metavar": "X"},
                    },
                },
            )
            y: GrandChild

        @dataclasses.dataclass
        class Parent:
            x: str
            y: Child
            z: str = "default"

        option_infos = cls_utils.Dataclasses.iter_fields(Parent)
        assert option_infos == (
            types.OptionInfo(
                parent_cls=Parent,
                path="x",
                cls=str,
                default=dataclasses.MISSING,
                has_no_default=True,
                default_is_factory=False,
            ),
            types.OptionInfo(
                parent_cls=Child,
                path="y.x",
                cls=float,
                default=dataclasses.MISSING,
                has_no_default=True,
                default_is_factory=False,
                metadata={
                    "click": {"help": "child x", "metavar": "X"},
                },
            ),
            types.OptionInfo(
                parent_cls=GrandChild,
                path="y.y.x",
                cls=Optional[int],  # type: ignore[arg-type]
                default=None,
                has_no_default=False,
                default_is_factory=False,
                metadata={
                    "argparse": {"help": "grand child x", "metavar": "GX"},
                    "click": {"help": "grand child x", "metavar": "GX"},
                    "help": "grand child x",
                },
            ),
            types.OptionInfo(
                parent_cls=Parent,
                path="z",
                cls=str,
                default="default",
                has_no_default=False,
                default_is_factory=False,
            ),
        )

    def test_unresolved_types(self) -> None:
        """Raise a NameError when types cannot be resolved."""

        @dataclasses.dataclass
        class C:
            name: str
            x: "X"  # type: ignore  # noqa: F821

        with pytest.raises(NameError, match="name 'X' is not defined"):
            cls_utils.Dataclasses.iter_fields(C)

    def test_direct_recursion(self) -> None:
        """
        We do not (and cannot easily) detect recursion.  A NameError is already
        raised when we try to resolve all types.  This is good enough.
        """

        @dataclasses.dataclass
        class Node:
            name: str
            child: "Node"

        with pytest.raises(NameError, match="name 'Node' is not defined"):
            cls_utils.Dataclasses.iter_fields(Node)

    def test_indirect_recursion(self) -> None:
        """
        We cannot (easily) detect indirect recursion but it is an error
        nonetheless.  This is not Dark!
        """

        @dataclasses.dataclass
        class Child:
            name: str
            parent: "Parent"

        @dataclasses.dataclass
        class Parent:
            name: str
            child: "Child"

        with pytest.raises(NameError, match="name 'Child' is not defined"):
            cls_utils.Dataclasses.iter_fields(Parent)

    def test_no_init_no_option(self) -> None:
        """
        No option is generated for an attribute if "init=False".
        """

        @dataclasses.dataclass(frozen=True)
        class Nested1:
            a: int = 0
            nb1: int = dataclasses.field(default=0, init=False)

        @dataclasses.dataclass(frozen=True)
        class Nested2:
            a: int = 0
            nb2: int = dataclasses.field(default=0, init=False)

        @dataclasses.dataclass
        class Settings:
            a: int = 0
            na: int = dataclasses.field(init=False)
            n1: Nested1 = Nested1()  # noqa: RUF009
            n2: Nested2 = Nested2()  # noqa: RUF009

        options = [o.path for o in cls_utils.Dataclasses.iter_fields(Settings)]
        assert options == ["a", "n1.a", "n2.a"]

    def test_fields_to_parent_classes(self) -> None:
        """
        If there are only scalar settings, create s single group.
        """

        @dataclasses.dataclass
        class Child1:
            x: int

        @dataclasses.dataclass
        class Child2:
            x: float

        @dataclasses.dataclass
        class Parent:
            a: str
            b: Child1
            c: Child2
            d: int

        result = cls_utils.Dataclasses.fields_to_parent_classes(Parent)
        assert result == {
            "a": Parent,
            "b": Child1,
            "c": Child2,
            "d": Parent,
        }


class TestPydantic:
    """Tests for Pydantic classes."""

    def test_check_true(self) -> None:
        """
        "check()" detects dataclasses.
        """

        class C(pydantic.BaseModel):
            x: int

        assert cls_utils.Pydantic.check(C)

    def test_check_false(self) -> None:
        """
        "check()" only detects dataclasses.
        """

        class C:
            x: int

        assert not cls_utils.Pydantic.check(C)

    def test_check_not_installed(self, unimport: Callable[[str], None]) -> None:
        """
        "check()" returns ``False`` if Pydantic is not installed.
        """

        class C(pydantic.BaseModel):
            x: int

        unimport("pydantic")

        assert not cls_utils.Pydantic.check(C)

    def test_iter_fields(self) -> None:
        """
        "iter_fields()" yields an option info for all options, including nested
        classes.
        """
        from pydantic_core._pydantic_core import PydanticUndefined

        class GrandChild(pydantic.BaseModel):
            x: Optional[int] = pydantic.Field(
                default=None,
                description="grand child x",
                json_schema_extra={
                    "typed-settings": {
                        "argparse": {"metavar": "GX"},
                        "click": {"metavar": "GX"},
                    },
                },
            )

        class Child(pydantic.BaseModel):
            x: "float" = pydantic.Field(  # Test resolving types
                json_schema_extra={
                    "typed-settings": {
                        "help": "child x",
                        "click": {"metavar": "X"},
                    },
                }
            )
            y: GrandChild

        class Parent(pydantic.BaseModel):
            x: str
            y: Child
            z: str = "default"

        option_infos = cls_utils.Pydantic.iter_fields(Parent)
        assert option_infos == (
            types.OptionInfo(
                parent_cls=Parent,
                path="x",
                cls=str,
                default=PydanticUndefined,
                has_no_default=True,
                default_is_factory=False,
            ),
            types.OptionInfo(
                parent_cls=Child,
                path="y.x",
                cls=float,
                default=PydanticUndefined,
                has_no_default=True,
                default_is_factory=False,
                metadata={
                    "argparse": {"help": "child x"},
                    "click": {"help": "child x", "metavar": "X"},
                    "help": "child x",
                },
            ),
            types.OptionInfo(
                parent_cls=GrandChild,
                path="y.y.x",
                cls=Optional[int],  # type: ignore[arg-type]
                default=None,
                has_no_default=False,
                default_is_factory=False,
                metadata={
                    "argparse": {"help": "grand child x", "metavar": "GX"},
                    "click": {"help": "grand child x", "metavar": "GX"},
                },
            ),
            types.OptionInfo(
                parent_cls=Parent,
                path="z",
                cls=str,
                default="default",
                has_no_default=False,
                default_is_factory=False,
            ),
        )

    @pytest.mark.skip(reason="Types are not resolved for Pydantic")
    def test_unresolved_types(self) -> None:  # pragma: no cover
        """Raise a NameError when types cannot be resolved."""

        class C(pydantic.BaseModel):
            name: str
            x: "X"  # type: ignore  # noqa: F821

        with pytest.raises(NameError, match="name 'X' is not defined"):
            cls_utils.Pydantic.iter_fields(C)

    @pytest.mark.skip(reason="RecursionError")
    def test_direct_recursion(self) -> None:  # pragma: no cover
        """
        We do not (and cannot easily) detect recursion.  A NameError is already
        raised when we try to resolve all types.  This is good enough.
        """

        class Node(pydantic.BaseModel):
            name: str
            child: "Node"

        with pytest.raises(NameError, match="name 'Node' is not defined"):
            cls_utils.Pydantic.iter_fields(Node)

    @pytest.mark.skip(reason="RecursionError")
    def test_indirect_recursion(self) -> None:  # pragma: no cover
        """
        We cannot (easily) detect indirect recursion but it is an error
        nonetheless.  This is not Dark!
        """

        class Child(pydantic.BaseModel):
            name: str
            parent: "Parent"

        class Parent(pydantic.BaseModel):
            name: str
            child: "Child"

        with pytest.raises(NameError, match="name 'Child' is not defined"):
            cls_utils.Pydantic.iter_fields(Parent)

    def test_fields_to_parent_classes(self) -> None:
        """
        If there are only scalar settings, create s single group.
        """

        class Child1(pydantic.BaseModel):
            x: int

        class Child2(pydantic.BaseModel):
            x: float

        class Parent(pydantic.BaseModel):
            a: str
            b: Child1
            c: Child2
            d: int

        result = cls_utils.Pydantic.fields_to_parent_classes(Parent)
        assert result == {
            "a": Parent,
            "b": Child1,
            "c": Child2,
            "d": Parent,
        }



"""
Framework agnostic utilities for generating CLI options from Typed Settings
options.
"""

from collections.abc import (
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
)
from typing import Protocol, Union, get_args, get_origin

from ._compat import PY_310


if PY_310:
    from types import UnionType
else:
    from typing import Union as UnionType  # type: ignore

from typing import Any, Collection, Dict, List, Optional, Tuple

from . import _core, converters, types


__all__ = [
    "NO_DEFAULT",
    "NoDefaultType",
    "TypeHandlerFunc",
    "TypeHandler",
    "TypeArgsMaker",
    "get_default",
    "check_if_optional",
]


class NoDefaultType:
    """
    Sentinel class to indicate the lack of a default value for an option when ``None``
    is ambiguous.

    ``NoDefaultType`` is a singleton. There is only ever one of it.
    """

    _singleton = None

    def __new__(cls) -> "NoDefaultType":
        if NoDefaultType._singleton is None:
            NoDefaultType._singleton = super().__new__(cls)
        return NoDefaultType._singleton

    def __repr__(self) -> str:
        return "NO_DEFAULT"


NO_DEFAULT = NoDefaultType()
"""
Sentinel that indicates the lack of a default value for an option.
"""


class DefaultFactorySentinel:
    def __repr__(self) -> str:
        return "(dynamic)"

    def __call__(self) -> object:
        return None


Default = Union[Any, None, NoDefaultType]
NoneType = type(None)
StrDict = Dict[str, Any]


class TypeHandlerFunc(Protocol):
    """
    **Protocol:** A function that returns keyword arguments for a CLI option
    for a specific type.
    """

    def __call__(self, type: type, default: Default, is_optional: bool) -> StrDict:
        """
        Return keyword arguments for creating an option for *type*.

        Args:
            type: The type to create the option for.
            default: The default value for the option.  May be ``None`` or
                :data:`NO_DEFAULT`.
            is_optional: Whether the original type was an
                :class:`~typing.Optional`.
        """
        ...


class TypeHandler(Protocol):
    """
    **Protocol:** Callbacks for the :class:`TypeArgsMaker` that provide
    framework specific arguments for various classes of CLI options.

    .. versionadded:: 2.0.0
    """

    def get_scalar_handlers(self) -> Dict[type, TypeHandlerFunc]:
        """
        Return a dict that maps specialized handlers for certain types (e.g.,
        enums or datetimes.

        Such a handler can look like this:

        .. code-block:: python

            def handle_mytype(
                type: type,
                default: Default,
                is_optional: bool,
            ) -> Dict[str, Any]:
                kwargs = {
                    "type": MyCliType(...)
                }
                if default not in (None, NO_DEFAULT):
                    kwargs["default"] = default.stringify()
                elif is_optional:
                    kwargs["default"] = None
                return kwargs

        Return:
            A dict mapping types to the corresponding type handler function.
        """
        ...

    def handle_scalar(
        self, type: Optional[type], default: Default, is_optional: bool
    ) -> StrDict:
        """
        Handle all scalars for which :func:`get_scalar_handlers()` does not
        provide a specific handler.

        Args:
            type: The type to create an option for.  Can be none if the option
                is untyped.
            default: The default value for the option. My be ``None`` or
                :data:`NO_DEFAULT`.
            is_optional: Whether or not the option type was marked as option
                or not.

        Return:
            A dictionary with keyword arguments for creating an option for the
            given type.
        """
        ...

    def handle_tuple(
        self,
        type_args_maker: "TypeArgsMaker",
        types: Tuple[Any, ...],
        default: Optional[Tuple],
        is_optional: bool,
    ) -> StrDict:
        """
        Handle options for structured tuples (i.e., not list-like tuples).

        Args:
            type_args_maker: The :class:`TypeArgsMaker` that called this
                function.
            types: The types of all tuple items.
            default: Either a tuple of default values or ``None``.
            is_optional: Whether or not the option type was marked as option
                or not.

        Return:
            A dictionary with keyword arguments for creating an option for the
            tuple.
        """
        ...

    def handle_collection(
        self,
        type_args_maker: "TypeArgsMaker",
        types: Tuple[Any, ...],
        default: Optional[List[Any]],
        is_optional: bool,
    ) -> StrDict:
        """
        Handle collections, add options to allow multiple values and to
        collect them in a list/collection.

        Args:
            type_args_maker: The :class:`TypeArgsMaker` that called this
                function.
            types: The types of the list items.
            default: Either a collection of default values or ``None``.
            is_optional: Whether or not the option type was marked as option
                or not.

        Return:
            A dictionary with keyword arguments for creating an option for the
            list type.
        """
        ...

    def handle_mapping(
        self,
        type_args_maker: "TypeArgsMaker",
        types: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        """
        Handle dictionaries.

        Args:
            type_args_maker: The :class:`TypeArgsMaker` that called this
                function.
            types: The types of keys and values.
            default: Either a mapping of default values, ``None`` or :data:`NO_DEFAULT`.
            is_optional: Whether or not the option type was marked as option
                or not.

        Return:
            A dictionary with keyword arguments for creating an option for the
            tuple.
        """
        ...


class TypeArgsMaker:
    """
    This class derives type information (in the form of keyword arguments)
    for CLI options from attributes of a settings class.

    For example, it could return a dict ``{"type": int, "default": 3}`` for
    an option ``val: int = 3``.

    It is agnostic of the CLI framework being used.  The specifics for each
    framework are implemented in a :class:`TypeHandler` that is passed to this
    class.

    The TypeArgsMaker differentitates between scalar and collection types
    (e.g., :samp:`int` vs. :samp:`list[int]`. It inspects each option (field)
    of a settings class and calls the appropriate method of the
    :class:`TypeHandler`:

    - If a type is in the dict returned by
      :meth:`TypeHandler.get_scalar_handlers()`, call the corresponding
      handler.

    - For other scalar types, call :meth:`TypeHandler.handle_scalar()`.

    - For structured tuples, call :meth:`TypeHandler.handle_tuple()`.

    - For collections (e.g., lists, sets, and list-like tuples), call
      :meth:`TypeHandler.handle_collection()`.

    - For mappings (e.g., dicts), call :meth:`TypeHandler.handle_mapping()`.

    .. versionchanged:: 2.0.0
       Complete refactoring and renamed from *TypeHandler* to *TypeArgsMaker*.
    """

    def __init__(
        self,
        type_handler: TypeHandler,
    ) -> None:
        self.type_handler = type_handler
        self.list_types = (
            list,
            Sequence,
            MutableSequence,
            set,
            frozenset,
            MutableSet,
        )
        self.tuple_types = (tuple,)
        self.mapping_types = (
            dict,
            Mapping,
            MutableMapping,
        )

    def get_kwargs(self, otype: Any, default: Default) -> StrDict:
        """
        Analyse the option type and return keyword arguments for creating a
        CLI option for it.

        Args:
            otype: The option's type.  It can be None if the user uses an
                untyped class.
            default: The default value for the option.  It can be anything, but the
                values ``None`` (possible default for optional types) and
                :data:`NO_DEFAULT` (no default set) should be handled explicitly.

        Return:
            A dictionary with keyword arguments for creating a CLI option in
            for a given framework.

        Raise:
            TypeError: If the *otype* has an unsupported type (e.g., a union
                type).
        """
        origin = get_origin(otype)
        args = get_args(otype)
        otype, default, origin, args, is_optional = check_if_optional(
            otype, default, origin, args
        )

        if otype is None:
            return self._handle_scalar(otype, default, is_optional)

        elif origin is None:
            scalar_handlers = self.type_handler.get_scalar_handlers()
            for target_type, get_kwargs in scalar_handlers.items():
                if types.is_new_type(otype):
                    otype = otype.__supertype__
                if issubclass(otype, target_type):
                    return get_kwargs(otype, default, is_optional)

            return self._handle_scalar(otype, default, is_optional)

        else:
            if origin in self.list_types:
                return self._handle_collection(otype, args, default, is_optional)
            elif origin in self.tuple_types:
                return self._handle_tuple(otype, args, default, is_optional)
            elif origin in self.mapping_types:
                return self._handle_mapping(otype, args, default, is_optional)

            raise TypeError(f"Cannot create CLI option for: {otype}")

    def _handle_scalar(
        self,
        type: Optional[type],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        """
        Get kwargs for scalar types.
        """
        return self.type_handler.handle_scalar(type, default, is_optional)

    def _handle_tuple(
        self,
        type: type,
        args: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        """
        Get kwargs for tuples.

        Call :meth:`_handle_collection()` for list like tuples.
        """
        if len(args) == 2 and args[1] == ...:
            # "Immutable list" variant of tuple
            return self._handle_collection(type, args, default, is_optional)

        # "struct" variant of tuple

        default_val: Optional[Tuple]
        if isinstance(default, tuple):
            if not len(default) == len(args):
                raise TypeError(
                    f"Default value must be of len {len(args)}: {len(default)}"
                )
            kwargs = {"strict": True} if PY_310 else {}
            default_val = tuple(
                self.get_kwargs(a, d)["default"]
                for a, d in zip(args, default, **kwargs)
            )
        else:
            default_val = None

        kwargs = self.type_handler.handle_tuple(self, args, default_val, is_optional)
        return kwargs

    def _handle_collection(
        self,
        type: type,
        args: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        """
        Get kwargs for collections (e.g., lists or list-like tuples) of the
        same type.
        """
        if isinstance(default, Collection):
            # Call get_kwargs() to get proper default value formatting
            default = [self.get_kwargs(args[0], d)["default"] for d in default]
        else:
            default = None

        kwargs = self.type_handler.handle_collection(self, args, default, is_optional)
        return kwargs

    def _handle_mapping(
        self,
        type: type,
        args: Tuple[Any, ...],
        default: Default,
        is_optional: bool,
    ) -> StrDict:
        """
        Get kwargs for mapping types (e.g, dicts).
        """
        kwargs = self.type_handler.handle_mapping(self, args, default, is_optional)
        return kwargs


def get_default(
    option_info: types.OptionInfo,
    settings: types.MergedSettings,
    converter: converters.Converter,
) -> Any:
    """
    Return the proper default value for an attribute.

    If possible, the default is taken from loaded settings.  Else, use the
    field's default value.

    Args:
        option_info: The option description for the attribute.
        settings: A (nested) dict with the loaded settings.
        converter: The converter to be used.

    Return:
        The default value to be used for the option.  This can also be ``None``
        or a "nothing" value (e.g., :data:`attrs.NOTHING`).
    """
    default: Any

    if option_info.path in settings:
        value, meta = settings[option_info.path]
        try:
            default = _core.convert_value(option_info, value, meta, converter)
        except Exception as e:
            raise ValueError(
                f"Invalid default {value!r} for option {option_info.path!r} with "
                f"type {option_info.cls}: {e!r}"
            ) from e

    elif option_info.default_is_factory:
        # Use a fake factory function to indicate a dynamic default value, that is only
        # computed when the CLI is invoked (and not when the options are generated).
        default = DefaultFactorySentinel()

    elif option_info.has_no_default:
        default = NO_DEFAULT

    else:
        default = option_info.default

    return default


def check_if_optional(
    otype: Optional[type],
    default: Default,
    origin: Any,
    args: Tuple[Any, ...],
) -> Tuple[Optional[type], Any, Any, Tuple[Any, ...], bool]:
    """
    Check if *otype* is an optional (``Optional[...]`` or ``Union[None, ...]``) and
    return the actual type for it and a flag indicating the optionality.

    If it is optional and a default value is not set, use ``None`` as new default.

    Args:
        otype: The Python type of the option.
        default: The option's default value.
        origin: The generic origin as returned by :func:`typing.get_origin()`.
        args: The generic args as returned by :func:`typing.get_args()`.

    Return:
        A tuple *(otype, default, origin, args, is_optional)*:

        *otype:*
            is either the original or the unwrapped optional type.
        *default:*
            is the possibly updated default value.
        *origin:*
            is the possibly updated *origin* for the unwrapped *otype*.
        *args:*
            are the possibly updated *args* for the unwrapped *otype*.
        *is_optional:*
            indicates whether *otype* was an optional or not.
    """
    is_optional = origin in (Union, UnionType) and len(args) == 2 and NoneType in args
    if is_optional:
        if default is NO_DEFAULT:
            default = None

        # "idx" is the index of the not-NoneType:
        idx = (args.index(NoneType) + 1) % 2
        otype = args[idx]
        origin = get_origin(otype)
        args = get_args(otype)

    return otype, default, origin, args, is_optional



"""
This module contains the settings processors provided by Typed Settings and the
protocol specification that they must implement.
"""

import logging
import subprocess
from typing import TYPE_CHECKING, Any, Dict, Optional, Protocol

from .dict_utils import iter_settings, set_path
from .types import OptionList, SettingsClass, SettingsDict


__all__ = [
    "Protocol",
    "UrlHandler",
    "Processor",
    "UrlProcessor",
    "handle_raw",
    "handle_script",
    "handle_op",
    "FormatProcessor",
    "JinjaProcessor",
]


LOGGER = logging.getLogger("typed_settings")


class Processor(Protocol):
    """
    **Protocol** that settings processors must implement.

    Processors must be callables (e.g., functions) with the specified
    signature.

    .. versionadded:: 23.0.0
    """

    def __call__(
        self,
        settings_dict: SettingsDict,
        settings_cls: SettingsClass,
        options: OptionList,
    ) -> SettingsDict:
        """
        Modify or update values in *settings_dict* and return an updated
        version.

        You may modify settings_dict in place - you don't need to return a
        copy of it.

        You should not add additional keys.

        Args:
            settings_dict: The dict of loaded settings.  Values are not yet
                converted to the target type (e.g., ``int`` values loaded from
                an env var are still a string).
            settings_cls: The base settings class for all options.
            options: The list of available settings.

        Return:
            The updated settings dict.
        """
        ...


class UrlHandler(Protocol):
    """
    **Protocol** that handlers for :class:`UrlProcessor` must implement.

    Handlers must be callables (e.g., functions) with the specified signature.

    .. versionadded:: 23.0.0
    """

    def __call__(self, value: str, scheme: str) -> str:
        """
        Handle the URL resource *value* and return the result.

        Args:
            value: The URL without the scheme (the ``v`` in ``s://v``).
            scheme: The URL scheme (the ``s://` in ``s://v``).

        Return:
            The result of the operation.

        Raise:
            ValueError: If the URL is invalid or another error occurs while
                handling the URL.
        """
        ...


class UrlProcessor:
    """
    Modify values that match one of the configured URL schemes.

    Args:
        handlers: A dictionary mapping URL schemes to handler functions.

    .. versionadded:: 23.0.0
    """

    def __init__(self, handlers: Dict[str, UrlHandler]) -> None:
        self.handlers = handlers
        """
        Registered URL scheme handlers.

        You can modify this dict after an instance of this class has been
        created.
        """

    def __call__(
        self,
        settings_dict: SettingsDict,
        settings_cls: SettingsClass,
        options: OptionList,
    ) -> SettingsDict:
        """
        Modify or update values in *settings_dict* and return an updated
        version.

        You may modify settings_dict in place - you don't need to return a
        copy of it.

        You should not add additional keys.

        Args:
            settings_dict: The dict of loaded settings.  Values are not yet
                converted to the target type (e.g., ``int`` values loaded from
                an env var are still a string).
            settings_cls: The base settings class for all options.
            options: The list of available settings.

        Return:
            The updated settings dict.
        """
        for path, value in iter_settings(settings_dict, options):
            for scheme, handler in self.handlers.items():
                if isinstance(value, str) and value.startswith(scheme):
                    start_idx = len(scheme)
                    value = value[start_idx:]
                    value = handler(value, scheme)
                    set_path(settings_dict, path, value)
                    break  # Only process a value once!

        return settings_dict


def handle_raw(value: str, scheme: str) -> str:
    """
    **URL handler:** Return *value* unchanged.

    .. versionadded:: 23.0.0
    """
    return value


def handle_script(value: str, scheme: str) -> str:
    """
    **URL handler:** Run *value* as shell script and return its output.

    .. attention::

       This handler can run arbitrary shell scripts with the permissions of your
       application.
       Only use this handler if you only load configuration from trusted sources!

    .. versionadded:: 23.0.0
    """
    try:
        result = subprocess.run(  # noqa: S602
            value,
            shell=True,
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        msg = (
            f"Helper script failed: {scheme}{value}\n"
            f"EXIT CODE: {e.returncode}\n"
            f"STDOUT:\n{e.stdout}"
            f"STDERR:\n{e.stderr}"
        )
        raise ValueError(msg) from e


def handle_op(value: str, scheme: str) -> str:
    """
    **URL handler:** Retrieve the resource *value* from the `1Password CLI`_.

    You must must have installed it and set it up in order for this to work.

    .. _1Password CLI: https://developer.1password.com/docs/cli/

    .. versionadded:: 23.0.0
    """
    from . import _onepassword

    return _onepassword.get_resource(f"op://{value}")


class FormatProcessor:
    """
    Perform value interpolation / templating via Python format strings.

    Formatting is performed recursively as long as the value is a valid format
    string.

    No exceptions are raised.  If format strings are invalid or refer to
    non existing values, they are returned unchanged.

    .. versionadded:: 23.0.0
    """

    def __call__(
        self,
        settings_dict: SettingsDict,
        settings_cls: SettingsClass,
        options: OptionList,
    ) -> SettingsDict:
        """
        Invoke the processor to render all values in the settings dict.
        """
        for path, value in iter_settings(settings_dict, options):
            value = self._render(value, settings_dict)
            set_path(settings_dict, path, value)

        return settings_dict

    def _render(self, value: Any, settings_dict: SettingsDict) -> Any:
        """
        Recursively render *value*.
        """
        if not self._is_possibly_format_string(value):
            return value

        try:
            new_value = value.format(**settings_dict)
        except Exception:
            return value

        if new_value == value:
            return new_value

        value = self._render(new_value, settings_dict)
        return value

    def _is_possibly_format_string(self, value: Any) -> bool:
        """
        Guess if *value* may be format string.

        It only detectecs if *value* is certainly *not* a format string and
        returns ``False`` in that case.
        If it returns ``True``, it may or may not be a valid format string.
        """
        return isinstance(value, str) and "{" in value and "}" in value


if TYPE_CHECKING:
    import jinja2


class JinjaProcessor:
    """
    Perform value templating with Jinja__.

    __ https://palletsprojects.com/p/jinja/

    Rendering is performed recursively as long as the value is a valid Jinja
    template.

    No exceptions are raised.  If templates are invalid or refer to non
    existing values, they are returned unchanged.

    Raises:
        ModuleNotFoundError: If ``jinja2`` is not installed.

    .. versionadded:: 23.0.0
    """

    def __init__(self, environment: Optional["jinja2.Environment"] = None) -> None:
        try:
            import jinja2
        except ImportError as e:
            raise ModuleNotFoundError(
                "Module 'jinja2' not installed.  Please run "
                "'python -m pip install -U typed-settings[jinja]'"
            ) from e

        self._jinja2 = jinja2
        # autoescape must be False or recursive rendering will not work
        # properly.
        if environment is None:
            self._env = jinja2.Environment(autoescape=False)  # noqa: S701
        else:
            self._env = environment
            self._env.autoescape = False

    def __call__(
        self,
        settings_dict: SettingsDict,
        settings_cls: SettingsClass,
        options: OptionList,
    ) -> SettingsDict:
        """
        Invoke the processor to render all values in the settings dict.
        """
        for path, value in iter_settings(settings_dict, options):
            value = self.render(value, settings_dict)
            set_path(settings_dict, path, value)

        return settings_dict

    def render(self, value: Any, settings_dict: SettingsDict) -> Any:
        """
        Recursively render *value*.
        """
        if not self.is_possibly_template(value):
            return value

        try:
            template = self._env.from_string(value)
            value = template.render(**settings_dict)
        except self._jinja2.TemplateError:
            return value

        value = self.render(value, settings_dict)
        return value

    def is_possibly_template(self, value: Any) -> bool:
        """
        Guess if *value* may be format string.

        It only detectecs if *value* is certainly *not* a format string and
        returns ``False`` in that case.
        If it returns ``True``, it may or may not be a valid format string.
        """
        if isinstance(value, str):
            for marker in (
                self._env.block_start_string,
                self._env.variable_start_string,
                self._env.comment_start_string,
            ):
                if marker in value:
                    return True
        return False



"""
Configuration for Sphinx docs.
"""

from importlib.metadata import version as get_version


project = "Typed Settings"
author = "Stefan Scherfke"
copyright = "2020, Stefan Scherfke"
release = get_version("typed-settings")
version = ".".join(release.split(".")[0:2])


extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
]
myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "fieldlist",
    "smartquotes",
]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

nitpick_ignore = [
    ("py:class", "NewTypeLike"),
    ("py:class", "attr.AttrsInstance"),
    ("py:class", "attrs.AttrsInstance"),
    ("py:class", "pydantic.SecretBytes"),
    ("py:class", "pydantic.SecretStr"),
    ("py:class", "typed_settings.cli_click.F"),
    ("py:class", "typed_settings.types._Auto"),
]


html_theme = "furo"
html_theme_options = {
    # "logo_only": True,
    # "sidebar_hide_name": True,
    "light_css_variables": {
        "color-brand-primary": "#266DB4",
        "color-brand-content": "#266DB4",
    },
    "dark_css_variables": {
        "color-brand-primary": "#3186DC",
        "color-brand-content": "#3186DC",
    },
}
html_static_path = ["_static"]
html_css_files = [
    "custom.css",
]
html_logo = "_static/typed-settings-spacing.svg"
html_title = "Typed Settings"


# Autodoc
autodoc_member_order = "bysource"

# Copybutton
copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True

# Intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
    "cattrs": ("https://catt.rs/en/latest/", None),
    "click": ("https://click.palletsprojects.com/en/latest/", None),
    "click-option-group": (
        "https://click-option-group.readthedocs.io/en/latest/",
        None,
    ),
    "jinja": ("https://jinja.palletsprojects.com/en/latest/", None),
}


# Workaround for https://github.com/sphinx-doc/sphinx/issues/10785
# (Type aliases are not properly resolved, change "py:class" to "py:data")
TYPE_ALIASES = ["SettingsDict"]


def resolve_type_aliases(app, env, node, contnode):
    """Resolve :class: references to our type aliases as :attr: instead."""
    if (
        node["refdomain"] == "py"
        and node["reftype"] == "class"
        and node["reftarget"] in TYPE_ALIASES
    ):
        return app.env.get_domain("py").resolve_xref(
            env, node["refdoc"], app.builder, "data", node["reftarget"], node, contnode
        )


def setup(app):  # noqa: D103
    app.connect("missing-reference", resolve_type_aliases)



"""
Shared fixtures for all tests.
"""

import dataclasses
import os
import sys
from pathlib import Path
from typing import Callable, Dict, Tuple

import pytest

from typed_settings.cls_utils import deep_options
from typed_settings.types import OptionList


# Test with frozen settings.  If it works this way, it will also work with
# mutable settings but not necessarily the other way around.
@dataclasses.dataclass(frozen=True)
class Host:
    """Host settings."""

    name: str
    port: int


@dataclasses.dataclass(frozen=True)
class Settings:
    """Main settings."""

    host: Host
    url: str
    default: int = 3


SettingsClasses = Tuple[type, type]


SETTINGS_CLASSES: Dict[str, SettingsClasses] = {"dataclasses": (Settings, Host)}

try:
    import attrs

    @attrs.frozen
    class HostAttrs:
        """Host settings."""

        name: str
        port: int

    @attrs.frozen
    class SettingsAttrs:
        """Main settings."""

        host: HostAttrs
        url: str
        default: int = 3

    SETTINGS_CLASSES["attrs"] = (SettingsAttrs, HostAttrs)
except ImportError:
    # "attrs" is not available in the nox session "test_no_optionals"
    pass

try:
    import pydantic

    class HostPydantic(pydantic.BaseModel):
        """Host settings."""

        name: str
        port: int

    class SettingsPydantic(pydantic.BaseModel):
        """Main settings."""

        host: HostPydantic
        url: str
        default: int = 3

    SETTINGS_CLASSES["pydantic"] = (SettingsPydantic, HostPydantic)
except ImportError:
    # "pydantic" is not available in the nox session "test_no_optionals"
    pass


@pytest.fixture(params=list(SETTINGS_CLASSES))
def settings_clss(request: pytest.FixtureRequest) -> SettingsClasses:
    """
    Return an example settings class.
    """
    return SETTINGS_CLASSES[request.param]


@pytest.fixture
def options(settings_clss: SettingsClasses) -> OptionList:
    """
    Return the option list for the example settings class.
    """
    main, _host = settings_clss
    return deep_options(main)


@pytest.fixture
def mock_op(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Update ``PATH`` in ``os.environ`` to point to a mocked 1Password CLI.
    """
    here = Path(__file__).parent
    path = os.getenv("PATH")
    path = f"{here}:{path}"
    monkeypatch.setitem(os.environ, "PATH", path)


@pytest.fixture
def unimport(monkeypatch: pytest.MonkeyPatch) -> Callable[[str], None]:
    """
    Return a function for unimporting modules and preventing reimport.

    Needed to test optional dependencies.
    """

    def unimport_module(modname: str) -> None:
        # Remove if already imported
        monkeypatch.delitem(sys.modules, modname, raising=False)
        # Prevent import:
        monkeypatch.setattr(sys, "path", [])

    return unimport_module



"""
Tests for "typed_settings.onepasword".
"""

import os
import subprocess
from pathlib import Path
from typing import Any, Optional, Tuple

import pytest
from packaging.version import Version

from typed_settings import _onepassword as op


HERE = Path(__file__).parent


@pytest.fixture(autouse=True, params=[False, True])
def mock_op_cli(
    request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    Generate two test runs for each test: One with the real "op" ClI and one with a
    mock.
    """
    if request.param:
        request.getfixturevalue("mock_op")
    else:
        try:  # pragma: no cover
            has_op = op.run("account", "list") != ""
        except ValueError:  # pragma: no cover
            has_op = False

        in_ci = "CI" in os.environ
        on_feature_branch = os.getenv("CI_COMMIT_BRANCH", "") not in {"main", ""}

        if (not has_op) or (in_ci and on_feature_branch):  # pragma: no cover
            pytest.skip(reason="OP not installed or credentials not accessible")


def test_op_run() -> None:
    """
    "run()" invokes the CLI with the provided arguments and returns its stdout.
    """
    result = op.run("--version")
    assert Version(result)


def test_op_not_installed(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    A helpful error is raised if op is not installed.
    """
    orig_run = subprocess.run

    def fake_run(cmd: Tuple[str, ...], **kwargs: Any):
        cmd = ("xyz",) + cmd[1:]
        return orig_run(cmd, **kwargs)

    monkeypatch.setattr(subprocess, "run", fake_run)
    msg = (
        "The 1Password CLI is not properly installed.*"
        "https://developer.1password.com/docs/cli"
    )
    with pytest.raises(ValueError, match=msg):
        op.run("--version")


def test_op_error() -> None:
    """
    An error is raised if the "op" invocation fails.
    """
    msg = '"op" error:.*(unknown command)|(No such command).*'
    with pytest.raises(ValueError, match=msg):
        op.run("spam", "eggs")


@pytest.mark.parametrize("vault", ["Test", None])
def test_get_item(vault: Optional[str]) -> None:
    """
    An item can be retrieved from 1Password.  Item labels are converted to
    dict keys.
    """
    result = op.get_item("Test", vault)
    assert result == {"username": "spam", "password": "eggs"}


def test_get_item_not_exists() -> None:
    """
    A ValueError with the "op" output is raised if an item does not exist.
    """
    msg = '"op" error: "xyz" isn\'t an item.*'
    with pytest.raises(ValueError, match=msg):
        op.get_item("xyz")


def test_get_resource() -> None:
    """
    A resource canbe retrieved from 1Password.
    """
    result = op.get_resource("op://Test/Test/password")
    assert result == "eggs"


def test_get_resource_not_exists() -> None:
    """
    A ValueError with the "op" output is raised if a resource does not exist.
    """
    msg = "\"op\" error: could not read secret 'op://Test/x': invalid secret reference"
    with pytest.raises(ValueError, match=msg):
        op.get_resource("op://Test/x")



"""
Deprecated alias for :mod:`typed_settings.cli_utils`.
"""

import warnings

from .cls_attrs import *  # noqa


warnings.warn(
    "This module has been renamed to 'typed_settings.cls_attrs'.",
    DeprecationWarning,
    stacklevel=2,
)



"""
Tests for `typed_settings.attrs.converters`.
"""

import dataclasses
import json
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from pathlib import Path
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import attrs
import cattrs
import pydantic
import pytest
from hypothesis import assume, given, strategies

from typed_settings import converters
from typed_settings._compat import PY_39, PY_310, PY_311
from typed_settings.cls_attrs import option, secret, settings


def custom_converter(v: Union[str, Path]) -> Path:
    """A custom converter for attrs fields."""
    return Path(v).resolve()


class LeEnum(Enum):
    """A simple enum for testing."""

    spam = "Le Spam"
    eggs = "Le Eggs"


@dataclasses.dataclass
class DataCls:
    """A basic "dataclass" for testing."""

    u: str
    p: str


@settings
class AttrsCls:
    """A basic "attrs" class for testing."""

    u: str = option()
    p: str = secret()


class PydanticCls(pydantic.BaseModel):
    """A basic Pydantic class."""

    u: str
    p: str


@dataclasses.dataclass
class ChildDc:
    """A simple nested class."""

    x: int
    y: Path


@dataclasses.dataclass(frozen=True)
class ParentDc:
    """A rather complex class with various scalar and composite attribute types."""

    child: ChildDc
    a: float
    c: LeEnum
    d: datetime
    e: List[ChildDc]
    f: Set[datetime]
    b: float = dataclasses.field(default=3.14)


@attrs.frozen
class ChildAttrs:
    """A simple nested class."""

    x: int
    y: Path = attrs.field(converter=custom_converter)


@attrs.frozen(kw_only=True)
class ParentAttrs:
    """A rather complex class with various scalar and composite attribute types."""

    child: ChildAttrs
    a: float
    b: float = attrs.field(default=3.14, validator=attrs.validators.le(2))
    c: LeEnum
    d: datetime
    e: List[ChildAttrs]
    f: Set[datetime]


class ChildPydantic(pydantic.BaseModel):
    """A simple nested class."""

    x: int
    y: Path


class ParentPydantic(pydantic.BaseModel):
    """A rather complex class with various scalar and composite attribute types."""

    child: ChildPydantic
    a: float
    b: float = pydantic.Field(default=3.14, le=4)
    c: LeEnum
    d: datetime
    e: List[ChildPydantic]
    f: Set[datetime]
    g: pydantic.SecretStr = pydantic.Field(default=pydantic.SecretStr("secret-default"))


Example3T = List[Tuple[str, Any, Any]]  # 3-tuple example
Example4T = List[Tuple[str, Any, Any, Any]]  # 4-tuple example

# This list is filled with examples for each supported data type below.
# It is used to check that all supported converters can convert the same data.
SUPPORTED_TYPES_DATA: Example4T = []

# Any - types remain unchanged
SUPPORTED_ANY: Example3T = [
    ("Any(int)", 2, 2),
    ("Any(str)", "2", "2"),
    ("Any(None)", None, None),
]
SUPPORTED_TYPES_DATA += [(n, v, e, Any) for n, v, e in SUPPORTED_ANY]

# bool - can be parsed from a defined set of values
SUPPORTED_BOOL: Example3T = [
    ("bool(True)", True, True),
    ("bool('True')", "True", True),
    ("bool('True')", "TRUE", True),
    ("bool('true')", "true", True),
    ("bool('true')", "t", True),
    ("bool('yes')", "yes", True),
    ("bool('yes')", "Y", True),
    ("bool('yes')", "on", True),
    ("bool('1')", "1", True),
    ("bool(1)", 1, True),
    ("bool(False)", False, False),
    ("bool('False')", "False", False),
    ("bool('False')", "fAlse", False),  # sic!
    ("bool('false')", "false", False),
    ("bool('no')", "NO", False),
    ("bool('no')", "n", False),
    ("bool('no')", "OFF", False),
    ("bool('0')", "0", False),
    ("bool(0)", 0, False),
]
SUPPORTED_TYPES_DATA += [(n, v, e, bool) for n, v, e in SUPPORTED_BOOL]

# int, float, str - nothing special about these ...
SUPPORTED_STDTYPES: Example4T = [
    # Nothing special about these ...
    ("int(23)", 23, 23, int),
    ("int('42')", "42", 42, int),
    ("float(3.14)", 3.14, 3.14, float),
    ("float('.815')", ".815", 0.815, float),
    ("str('spam')", "spam", "spam", str),
]
SUPPORTED_TYPES_DATA += SUPPORTED_STDTYPES

# datetime - can be parsed from ISO format
SUPPORTED_DATETIME: Example3T = [
    ("datetime(naive-space)", "2020-05-04 13:37:00", datetime(2020, 5, 4, 13, 37)),
    ("datetime(naive-T)", "2020-05-04T13:37:00", datetime(2020, 5, 4, 13, 37)),
    (
        "datetime(tz-Z)",
        "2020-05-04T13:37:00Z",
        datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
    ),
    (
        "datetime(tz-offset-utc)",
        "2020-05-04T13:37:00+00:00",
        datetime(2020, 5, 4, 13, 37, tzinfo=timezone.utc),
    ),
    (
        "datetime(tz-offset-2h)",
        "2020-05-04T13:37:00+02:00",
        datetime(2020, 5, 4, 13, 37, tzinfo=timezone(timedelta(seconds=7200))),
    ),
    ("datetime(inst)", datetime(2020, 5, 4, 13, 37), datetime(2020, 5, 4, 13, 37)),
]
SUPPORTED_TYPES_DATA += [(n, v, e, datetime) for n, v, e in SUPPORTED_DATETIME]

# date - can be parsed from ISO format
SUPPORTED_DATE: Example3T = [
    ("date(str)", "2020-05-04", date(2020, 5, 4)),
    ("date(str-no-dashes)", "20200504" if PY_311 else "2020-05-04", date(2020, 5, 4)),
    ("date(inst)", date(2020, 5, 4), date(2020, 5, 4)),
]
SUPPORTED_TYPES_DATA += [(n, v, e, date) for n, v, e in SUPPORTED_DATE]

# timedelta - can be parsed from ISO and simpliefied string formats
SUPPORTED_TIMEDELTA: Example3T = [
    (f"timedelta({kind}-{example})", td_str, td)
    for (td, example), examples in {
        (timedelta(seconds=1), "s"): [
            ("iso", "PT01S"),
            ("simple-iso", "1s"),
            ("simple", "1"),
        ],
        (timedelta(minutes=1, seconds=1), "ms"): [
            ("iso", "PT01M01S"),
            ("simple-iso", "1m1s"),
            ("simple", "1:01"),
        ],
        (timedelta(hours=1, minutes=1, seconds=1), "hms"): [
            ("iso", "PT01H01M01S"),
            ("simple-iso", "1h1m1s"),
            ("simple", "1:01:01"),
        ],
        (timedelta(days=1, hours=1, minutes=1, seconds=1), "dhms"): [
            ("iso", "P01DT01H01M01S"),
            ("simple-iso", "1d1h1m1s"),
            ("simple", "1d,1:01:01"),
        ],
        (timedelta(days=1, seconds=1), "ds"): [
            ("iso", "P1DT1S"),
            ("simple-iso", "1d1s"),
            ("simple", "1d1"),
        ],
        (timedelta(hours=1, seconds=1), "hs"): [
            ("iso", "PT01H01S"),
            ("simple-iso", "1h1s"),
            ("simple", "1:0:01"),
        ],
        (timedelta(seconds=-(24 * 60 * 60 + 1)), "-ds"): [
            ("iso", "-P01DT01S"),
            ("simple-iso", "-1d1s"),
            ("simple", "-1D00:00:01"),
        ],
        (timedelta(microseconds=1), "mic1"): [
            ("iso", "PT0.000001S"),
            ("simple-iso", "0.000001s"),
            ("simple", "0.000001"),
        ],
        (timedelta(microseconds=100_000), "mic1hk"): [
            ("iso", "PT0.1S"),
            ("simple-iso", "0.1s"),
            ("simple", "0.1"),
        ],
    }.items()
    for kind, td_str in examples
] + [
    ("timedelta(float)", 1.0, timedelta(seconds=1)),
    ("timedelta(inst)", timedelta(days=1, seconds=2), timedelta(days=1, seconds=2)),
]
SUPPORTED_TYPES_DATA += [(n, v, e, timedelta) for n, v, e in SUPPORTED_TIMEDELTA]

# Enum - Enums are parsed from their "key"
SUPPORTED_ENUM: Example3T = [
    ("enum(str)", "eggs", LeEnum.eggs),
    ("enum(inst)", LeEnum.eggs, LeEnum.eggs),
]
SUPPORTED_TYPES_DATA += [(n, v, e, LeEnum) for n, v, e in SUPPORTED_ENUM]

# Path - Paths are resolved by default
SUPPORTED_PATH = [
    ("path(str)", "spam", Path.cwd().joinpath("spam")),
    ("path(inst)", Path("eggs"), Path.cwd().joinpath("eggs")),
]
SUPPORTED_TYPES_DATA += [(n, v, e, Path) for n, v, e in SUPPORTED_PATH]

# Pydantic Secret Str|Bytes
SUPPORTED_PYDANTIC_SECRET = [
    ("pydantic.SecretStr", "x", pydantic.SecretStr("x"), pydantic.SecretStr),
    ("pydantic.SecretBytes", b"x", pydantic.SecretBytes(b"x"), pydantic.SecretBytes),
    (
        "pydantic.SecretStr",
        pydantic.SecretStr("x"),
        pydantic.SecretStr("x"),
        pydantic.SecretStr,
    ),
    (
        "pydantic.SecretBytes",
        pydantic.SecretBytes(b"x"),
        pydantic.SecretBytes(b"x"),
        pydantic.SecretBytes,
    ),
]
SUPPORTED_TYPES_DATA += SUPPORTED_PYDANTIC_SECRET

# list
SUPPORTED_LIST: Example4T = [
    ("List[any]", [1, "2"], [1, "2"], List),
    ("list[any]", [1, "2"], [1, "2"], list),
    ("List[int]", [1, 2], [1, 2], List[int]),
    (
        "list[datetime]",
        ["2023-05-04T13:37:42+00:00"],
        [datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)],
        List[datetime],
    ),
]
if PY_39:
    SUPPORTED_LIST += [
        ("list[int]", [1, "2"], [1, 2], list[int]),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_LIST

# tuple
SUPPORTED_TUPLE: Example4T = [
    ("tuple[any]", [1, "2"], (1, "2"), tuple),
    ("Tuple[Any]", [1, "2"], (1, "2"), Tuple),
    ("Tuple[int, ...]", [1, 2, "3"], (1, 2, 3), Tuple[int, ...]),
    ("Tuple[int, float]", [1, "2.3"], (1, 2.3), Tuple[int, float]),
    (
        "list[datetime]",
        ["2023-05-04T13:37:42+00:00"],
        (datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc),),
        Tuple[datetime],
    ),
]
if PY_39:
    SUPPORTED_TUPLE += [
        ("tuple[int, ...]", [1, 2, "3"], (1, 2, 3), tuple[int, ...]),
        ("tuple[int, float]", [1, "2.3"], (1, 2.3), tuple[int, float]),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_TUPLE

# dict
SUPPORTED_DICT: Example4T = [
    ("dict[any, any]", {"y": 1, "n": 3.1}, {"y": 1, "n": 3.1}, dict),
    ("Dict[Any, Any]", {"y": 1, "n": 3.1}, {"y": 1, "n": 3.1}, Dict),
    ("Dict[bool, int]", {"y": 1, "n": 3.1}, {True: 1, False: 3}, Dict[bool, int]),
    (
        "dict[str, datetime]",
        {"a": "2023-05-04T13:37:42+00:00"},
        {"a": datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)},
        Dict[str, datetime],
    ),
]
if PY_39:
    SUPPORTED_DICT += [
        (
            "dict[bool, int]",
            {"y": 1, "n": 3.1},
            {True: 1, False: 3},
            dict[bool, int],
        ),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_DICT

# MappingProxy
SUPPORTED_MAPPINGPROXY: Example4T = []
if PY_39:
    SUPPORTED_MAPPINGPROXY += [
        (
            "MappingProxyType[Any, Any]",
            {"y": 1, "n": 3.1},
            MappingProxyType({"y": 1, "n": 3.1}),
            MappingProxyType,
        ),
        (
            "MappingProxyType[bool, int]",
            {"y": 1, "n": 3.1},
            MappingProxyType({True: 1, False: 3}),
            MappingProxyType[bool, int],
        ),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_MAPPINGPROXY

# set
SUPPORTED_SET: Example4T = [
    ("set[any]", [1, "2"], {1, "2"}, set),
    ("Set[any]", [1, "2"], {1, "2"}, Set),
    ("Set[int]", [1, 2], {1, 2}, Set[int]),
]
if PY_39:
    SUPPORTED_SET += [
        ("set[int]", [1, "2"], {1, 2}, set[int]),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_SET

# frozenset
SUPPORTED_FROZENSET: Example4T = [
    ("frozenset[any]", [1, "2"], frozenset({1, "2"}), frozenset),
    ("Frozenset[any]", [1, "2"], frozenset({1, "2"}), FrozenSet),
    ("Frozenset[int]", [1, 2], frozenset({1, 2}), FrozenSet[int]),
]
if PY_39:
    SUPPORTED_FROZENSET += [
        ("frozenset(int)", [1, "2"], frozenset({1, 2}), frozenset[int]),
    ]
SUPPORTED_TYPES_DATA += SUPPORTED_FROZENSET

# Union / Optional
SUPPORTED_UNION: Example4T = [
    ("Optional(None)", None, None, Optional[str]),
    ("Optional(int)", 1, "1", Optional[str]),
    ("dc|None(None)", None, None, Optional[DataCls]),
    ("dc|None(dict)", {"u": "u", "p": "p"}, DataCls("u", "p"), Optional[DataCls]),
    ("enum|None", "spam", LeEnum.spam, Optional[LeEnum]),
    # ("Union(None)", None, None, Union[None, S, List[str]]),
    # (
    #     "Union(attrs)",
    #     {"u": "u", "p": "p"},
    #     S("u", "p"),
    #     Union[None, S, List[str]],
    # ),
    # ("Union(list)", [1, 2], ["1", "2"], Union[None, S, List[str]]),
]
SUPPORTED_TYPES_DATA += SUPPORTED_UNION
if PY_310:
    SUPPORTED_UNION = [
        ("str|None(None)", None, None, str | None),
        ("str|None(int)", 1, "1", str | None),
        # (S | List[str], [1, 2], ["1", "2"], "attrs|list(list)"),
    ]

# attrs classes
SUPPORTED_ATTRSCLASSES: Example4T = [
    ("attrs(dict)", {"u": "user", "p": "pwd"}, AttrsCls("user", "pwd"), AttrsCls),
    ("attrs(inst)", AttrsCls("user", "pwd"), AttrsCls("user", "pwd"), AttrsCls),
    (
        "attrs(nested)",
        {
            "a": "3.14",
            "b": 1,
            "c": "eggs",
            "d": "2023-05-04T13:37:42+00:00",
            "e": [{"x": 0, "y": "a"}, {"x": 1, "y": "b"}],
            "f": ["2023-05-04T13:37:42+00:00", "2023-05-04T13:37:42+00:00"],
            "child": {"x": 3, "y": "c"},
        },
        ParentAttrs(
            a=3.14,
            b=1,
            c=LeEnum.eggs,
            d=datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc),
            e=[
                ChildAttrs(0, Path.cwd().joinpath("a")),
                ChildAttrs(1, Path.cwd().joinpath("b")),
            ],
            f={datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)},
            child=ChildAttrs(3, Path.cwd().joinpath("c")),
        ),
        ParentAttrs,
    ),
]
SUPPORTED_TYPES_DATA += SUPPORTED_ATTRSCLASSES

# dataclasses
SUPPORTED_DATACLASSES: Example4T = [
    ("dc(dict)", {"u": "user", "p": "pwd"}, DataCls("user", "pwd"), DataCls),
    ("dc(inst)", DataCls("user", "pwd"), DataCls("user", "pwd"), DataCls),
    (
        "dc(nested)",
        {
            "a": "3.14",
            "b": 1,
            "c": "eggs",
            "d": "2023-05-04T13:37:42+00:00",
            "e": [{"x": 0, "y": "a"}, {"x": 1, "y": "b"}],
            "f": ["2023-05-04T13:37:42+00:00", "2023-05-04T13:37:42+00:00"],
            "child": {"x": 3, "y": "c"},
        },
        ParentDc(
            a=3.14,
            b=1,
            c=LeEnum.eggs,
            d=datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc),
            e=[
                ChildDc(0, Path.cwd().joinpath("a")),
                ChildDc(1, Path.cwd().joinpath("b")),
            ],
            f={datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)},
            child=ChildDc(3, Path.cwd().joinpath("c")),
        ),
        ParentDc,
    ),
]
SUPPORTED_TYPES_DATA += list(SUPPORTED_DATACLASSES)

# Pydantic classes
SUPPORTED_PYDANTIC: Example4T = [
    (
        "pydantic(dict)",
        {"u": "user", "p": "pwd"},
        PydanticCls(u="user", p="pwd"),
        PydanticCls,
    ),
    (
        "pydantic(inst)",
        PydanticCls(u="user", p="pwd"),
        PydanticCls(u="user", p="pwd"),
        PydanticCls,
    ),
    (
        "pydantic(nested)",
        {
            "a": "3.14",
            "b": 1,
            "c": "Le Eggs",
            "d": "2023-05-04T13:37:42+00:00",
            "e": [{"x": 0, "y": "a"}, {"x": 1, "y": "b"}],
            "f": ["2023-05-04T13:37:42+00:00", "2023-05-04T13:37:42+00:00"],
            "g": "secret-string",
            "child": {"x": 3, "y": "c"},
        },
        ParentPydantic(
            a=3.14,
            b=1,
            c=LeEnum.eggs,
            d=datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc),
            e=[
                ChildPydantic(x=0, y=Path("a")),
                ChildPydantic(x=1, y=Path("b")),
            ],
            f={datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)},
            g=pydantic.SecretStr("secret-string"),
            child=ChildPydantic(x=3, y=Path("c")),
        ),
        ParentPydantic,
    ),
    (
        "pydantic(nested) defaults",
        {
            "a": "3.14",
            "c": "Le Eggs",
            "d": "2023-05-04T13:37:42+00:00",
            "e": [{"x": 0, "y": "a"}, {"x": 1, "y": "b"}],
            "f": ["2023-05-04T13:37:42+00:00", "2023-05-04T13:37:42+00:00"],
            "child": {"x": 3, "y": "c"},
        },
        ParentPydantic(
            a=3.14,
            b=3.14,
            c=LeEnum.eggs,
            d=datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc),
            e=[
                ChildPydantic(x=0, y=Path("a")),
                ChildPydantic(x=1, y=Path("b")),
            ],
            f={datetime(2023, 5, 4, 13, 37, 42, tzinfo=timezone.utc)},
            g=pydantic.SecretStr("secret-default"),
            child=ChildPydantic(x=3, y=Path("c")),
        ),
        ParentPydantic,
    ),
]
SUPPORTED_TYPES_DATA += list(SUPPORTED_PYDANTIC)


@pytest.mark.parametrize(
    "converter",
    [
        pytest.param(converters.get_default_cattrs_converter(), id="converter:cattrs"),
        pytest.param(converters.get_default_ts_converter(), id="converter:ts"),
    ],
)
@pytest.mark.parametrize(
    "value, typ, expected",
    [pytest.param(v, t, e, id=n) for n, v, e, t in SUPPORTED_TYPES_DATA],
)
def test_supported_types(
    converter: converters.Converter, value: Any, typ: type, expected: Any
) -> None:
    """
    All officially supported types can be converted.

    The list :data:`SUPPORTED_TYPES_DATA` is the officially source of truth.

    Please create an issue if something is missing here.
    """
    assert converter.structure(value, typ) == expected


class TestToTimedelta:
    """
    Tests for "to_timedelta()".
    """

    @given(
        simple_fmt=strategies.booleans(),
        left_pad=strategies.booleans(),
        sign=strategies.sampled_from([None, "+", "-"]),
        days=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        hours=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        minutes=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        seconds=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        micros=strategies.one_of(strategies.none(), strategies.integers(0, 999999)),
    )
    def test_from_iso_string(
        self,
        simple_fmt: bool,
        left_pad: bool,
        sign: Optional[str],
        days: Optional[int],
        hours: Optional[int],
        minutes: Optional[int],
        seconds: Optional[int],
        micros: Optional[int],
    ) -> None:
        """
        Timedeltas can be parsed from ISO strings and simplified ISO strings (missing
        the "P" and the "T").

        - P[nnD][T[nnH][nnM][nnS]]
        - [nnD][nnH][nnM][nnS]

        The regex is case-insensitive.  Numbers may be left-padded with 0 (e.g., "02")
        and they can also be > 99.

        All places (days, hours, minutes, seconds) are optional.
        """
        # Ditch examples where everything is None
        assume(any(i is not None for i in [days, hours, minutes, seconds, micros]))

        if micros is not None and not seconds:
            seconds = 0

        fmt = ">02d" if left_pad else "d"

        # Strip trailing "0" from microseconds and make shure "x." -> "x.0"
        micros_str = "" if micros is None else f".{micros:>06}".rstrip("0")
        micros_str = ".0" if micros_str == "." else micros_str
        time_str = "" if hours is None else f"{hours:{fmt}}H"
        time_str += "" if minutes is None else f"{minutes:{fmt}}M"
        time_str += "" if seconds is None else f"{seconds:{fmt}}{micros_str}S"
        td_str = "" if sign is None else sign
        td_str += "P"
        td_str += "" if days is None else f"{days}D"
        td_str += f"T{time_str}" if time_str else ""
        if simple_fmt:
            td_str = td_str.replace("P", "").replace("T", "").lower()

        _kwargs = {
            "days": days,
            "hours": hours,
            "minutes": minutes,
            "seconds": seconds,
            "microseconds": micros,
        }
        kwargs = {
            k: (-v if sign == "-" else v) for k, v in _kwargs.items() if v is not None
        }
        assert converters.to_timedelta(td_str, timedelta) == timedelta(**kwargs), td_str

    @given(
        left_pad=strategies.booleans(),
        sign=strategies.sampled_from([None, "+", "-"]),
        comma=strategies.sampled_from(["", ","]),
        days=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        hours=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        minutes=strategies.one_of(strategies.none(), strategies.integers(0, 9999)),
        seconds=strategies.integers(0, 9999),
        micros=strategies.one_of(strategies.none(), strategies.integers(0, 999999)),
    )
    def test_from_simple_string(
        self,
        left_pad: bool,
        sign: Optional[str],
        comma: str,
        days: Optional[int],
        hours: Optional[int],
        minutes: Optional[int],
        seconds: int,
        micros: Optional[int],
    ) -> None:
        """
        Timedeltas can be parsed a simple string format..

        - [dD[,]][[hh:]mm:]ss[.ffffff]

        Numbers may be left-padded with 0 (e.g., "02") and they can also be > 99.
        """
        if hours is not None and minutes is None:
            minutes = 0
        fmt = ">02d" if left_pad else "d"
        micros_str = "" if micros is None else f".{micros:>06}".rstrip("0")
        micros_str = ".0" if micros_str == "." else micros_str
        td_str = f"{seconds:{fmt}}{micros_str}"
        td_str = td_str if minutes is None else f"{minutes:{fmt}}:{td_str}"
        td_str = td_str if hours is None else f"{hours:{fmt}}:{td_str}"
        td_str = td_str if days is None else f"{days}D{comma}{td_str}"
        td_str = td_str if sign is None else f"{sign}{td_str}"

        _kwargs = {
            "days": days,
            "hours": hours,
            "minutes": minutes,
            "seconds": seconds,
            "microseconds": micros,
        }
        kwargs = {
            k: (-v if sign == "-" else v) for k, v in _kwargs.items() if v is not None
        }
        assert converters.to_timedelta(td_str, timedelta) == timedelta(**kwargs), td_str

    @given(
        positive=strategies.booleans(),
        days=strategies.integers(0, 5),
        hours=strategies.integers(0, 23),
        minutes=strategies.integers(0, 59),
        seconds=strategies.integers(0, 59),
        micros=strategies.integers(0, 999999),
    )
    def test_timedelta_to_str(
        self,
        positive: bool,
        days: int,
        hours: int,
        minutes: int,
        seconds: int,
        micros: int,
    ) -> None:
        """
        Timedeltas can be converted back to a timedelta string (in simplified ISO
        format).
        """
        td = timedelta(
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            microseconds=micros,
        )
        if not positive:
            td = -td

        micros_str = f".{micros:>06}".rstrip("0") if micros else ""
        seconds_str = f"{seconds}{micros_str}s" if seconds or micros_str else ""
        expected = ""
        expected += f"{days}d" if days else ""
        expected += f"{hours}h" if hours else ""
        expected += f"{minutes}m" if minutes else ""
        expected += seconds_str
        if expected and not positive:
            expected = f"-{expected}"

        result = converters.timedelta_to_str(td)
        assert result == expected
        assert converters.to_timedelta(result, timedelta) == td


@pytest.mark.parametrize(
    "cls, value",
    [
        # "to_bool()" is flexible, but does not accept anything
        (bool, ""),
        (bool, []),
        (bool, "spam"),
        (bool, 2),
        (bool, -1),
        (datetime, 3),
        (date, 3),
        (timedelta, datetime(1, 1, 1)),
        (timedelta, "1s1h"),  # Not ordered properly
        # len(value) does not match len(tuple-args)
        (Tuple[int, int], (1,)),
        (Tuple[int, int], (1, 2, 3)),
        (Union[int, datetime, None], "3.1"),  # float is not part of the Union
        (Sequence, [0, 1]),  # Type not supported
        (AttrsCls, {"foo": 3}),  # Invalid attribute
        (AttrsCls, {"opt", "x"}),  # Invalid value
        (DataCls, {"foo": 3}),  # Invalid attribute
        (DataCls, {"opt", "x"}),  # Invalid value
        (PydanticCls, {"foo": 3}),  # Invalid attribute
        (PydanticCls, {"opt", "x"}),  # Invalid value
    ],
)
def test_unsupported_values(value: Any, cls: type) -> None:
    """
    Unsupported input leads to low level exceptions.  These are later unified by
    "_core.convert()".
    """
    converter = converters.TSConverter()
    with pytest.raises((KeyError, TypeError, ValueError)):
        converter.structure(value, cls)


@pytest.mark.parametrize(
    "converter",
    [
        pytest.param(converters.get_default_cattrs_converter, id="converter:cattrs"),
        pytest.param(converters.get_default_ts_converter, id="converter:ts"),
    ],
)
@pytest.mark.parametrize(
    "value, expected", [pytest.param(v, e, id=n) for n, v, e in SUPPORTED_PATH]
)
@pytest.mark.parametrize("resolve_paths", [True, False])
def test_resolve_path(
    converter: Callable[[bool], converters.Converter],
    value: Any,
    expected: Any,
    resolve_paths: bool,
) -> None:
    """
    The path-resolving behavior can be explicitly set.
    """
    if not resolve_paths:
        expected = expected.relative_to(Path.cwd())
    c = converter(resolve_paths)
    assert c.structure(value, Path) == expected


STRLIST_TEST_DATA = [
    (List[int], [1, 2, 3]),
    (Set[int], {1, 2, 3}),
    (FrozenSet[int], frozenset({1, 2, 3})),
    (Tuple[int, ...], (1, 2, 3)),
    (Tuple[int, int, int], (1, 2, 3)),
]

if PY_39:
    STRLIST_TEST_DATA.extend(
        [
            (list[int], [1, 2, 3]),
            (set[int], {1, 2, 3}),
            (tuple[int, ...], (1, 2, 3)),
        ]
    )


@pytest.mark.parametrize("cls_decorator", [attrs.frozen, dataclasses.dataclass])
@pytest.mark.parametrize(
    "input, kw", [("1:2:3", {"sep": ":"}), ("[1,2,3]", {"fn": json.loads})]
)
@pytest.mark.parametrize("typ, expected", STRLIST_TEST_DATA)
def test_cattrs_strlist_hook(
    cls_decorator: Callable, input: str, kw: dict, typ: type, expected: Any
) -> None:
    """
    The strlist hook for can be configured with a separator string or a function.
    """

    @cls_decorator
    class Settings:
        a: typ  # type: ignore

    converter = converters.get_default_cattrs_converter()
    converters.register_strlist_hook(converter, **kw)
    result = converter.structure({"a": input}, Settings)
    assert result == Settings(expected)  # type: ignore[call-arg]


def test_cattrs_strlist_hook_either_arg() -> None:
    """
    Either "sep" OR "fn" can be passed to "register_str_list_hook()".
    """
    converter = converters.get_default_cattrs_converter()
    with pytest.raises(ValueError, match="You may either pass"):
        converters.register_strlist_hook(
            converter, sep=":", fn=lambda v: [v]
        )  # pragma: no cover


@pytest.mark.parametrize("cls_decorator", [attrs.frozen, dataclasses.dataclass])
@pytest.mark.parametrize(
    "input, sep", [("1:2:3", ":"), ("[1,2,3]", json.loads), ("123", None)]
)
@pytest.mark.parametrize("typ, expected", STRLIST_TEST_DATA)
def test_ts_strlist_hook(
    cls_decorator: Callable,
    input: str,
    sep: Union[str, Callable],
    typ: type,
    expected: Any,
) -> None:
    """
    The TSConverter has a builtin strlist hook that takes a separator string or a
    function.  It can be disabled with ``None``.
    """

    @cls_decorator
    class Settings:
        a: typ  # type: ignore

    converter = converters.TSConverter(strlist_sep=sep)
    result = converter.structure({"a": input}, Settings)
    assert result == Settings(expected)  # type: ignore[call-arg]


def test_get_default_converter_cattrs_installed() -> None:
    """
    If cattrs is installed, a cattrs converter is used by default.
    """
    converter = converters.default_converter()
    assert isinstance(converter, cattrs.Converter)


def test_get_default_converter_cattrs_uninstalled(
    unimport: Callable[[str], None],
) -> None:
    """
    If cattrs is not installed, the builtin converter is used by default.
    """
    unimport("cattrs")
    converter = converters.default_converter()
    assert isinstance(converter, converters.TSConverter)


def test_get_cattrs_converter_uninstalled(unimport: Callable[[str], None]) -> None:
    """
    An exception is raised by "get_default_cattrs_converter()" if  cattrs is not
    installed.
    """
    unimport("cattrs")
    with pytest.raises(ModuleNotFoundError):
        converters.get_default_cattrs_converter()


@pytest.mark.parametrize(
    "get_converter",
    [
        pytest.param(converters.get_default_cattrs_converter, id="converter:cattrs"),
        pytest.param(converters.get_default_ts_converter, id="converter:ts"),
    ],
)
def test_pydantic_converters_pydantic_uninstalled(
    get_converter: Callable[[], converters.Converter], unimport: Callable[[str], None]
) -> None:
    """
    If pydantic is not installed, the Pydantic converters are not available.
    """
    unimport("pydantic")
    converter = get_converter()
    with pytest.raises((TypeError, cattrs.errors.StructureHandlerNotFoundError)):
        converter.structure("x", pydantic.SecretStr)
    with pytest.raises((TypeError, cattrs.errors.StructureHandlerNotFoundError)):
        converter.structure(b"x", pydantic.SecretBytes)



"""
Helpers and wrappers for settings class backends.

Supported backends are:

- :mod:`dataclasses`
- `attrs <https://attrs.org>`_ (optional dependency)
- `pydantic <https://docs.pydantic.dev>`_ (optional dependency)
"""

import dataclasses
import functools
from itertools import groupby
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from . import constants, types
from ._compat import PY_39


class ClsHandler(Protocol):
    """
    **Protocol** that class handlers must implement.

    .. versionadded:: 23.1.0
    """

    @staticmethod
    def check(cls: type) -> bool:
        """
        Return a bool indicating whether *cls* belongs to the handler's class lib.
        """

    @staticmethod
    def iter_fields(cls: type) -> types.OptionList:
        """
        Recursively iterate the the fields of *cls* and return the
        :class:`.types.OptionInfo` instances for them.

        Fields of nested classes are only converted to :class:`.types.OptionInfo` if
        they were created by the same class lib.  For example, if the parent class is
        an attrs class, the attributes of nested dataclasses are not added to the list
        of options.
        """

    @staticmethod
    def fields_to_parent_classes(cls: type) -> Dict[str, type]:
        """
        Map a class' attribute names to a "parent class".

        This parent class is used to create CLI option groups.  Thus, if a field's
        type is another (nested) settings class, that class should be used.  Else,
        the class itself should be used.
        """

    @staticmethod
    def asdict(inst: Any) -> types.SettingsDict:
        """
        Return the instances attributes as dict, recurse into nested classes of the
        same kind.
        """

    @staticmethod
    def resolve_types(
        cls: Type[types.T],
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        include_extras: bool = True,
    ) -> Type[types.T]:
        """
        Resolve any strings and forward annotations in type annotations.

        With no arguments, names will be looked up in the module in which the class was
        created.  If this is not what you want, e.g. if the name only exists inside
        a method, you may pass *globalns* or *localns* to specify other dictionaries in
        which to look up these names.  See the docs of :func:`typing.get_type_hints()`
        for more details.

        Args:
            cls: Class to resolve.
            globalns: Dictionary containing global variables.
            localns: Dictionary containing local variables.
            include_extras: Resolve more accurately, if possible.
                Pass ``include_extras`` to ``typing.get_hints``, if supported by the
                typing module.  On supported Python versions (3.9+), this resolves the
                types more accurately.

        Return: *cls* so you can use this function also as a class decorator.  Please
            note that you have to apply it **after** :func:`attrs.define` or
            :func:`dataclasses.dataclass`.  That means the decorator has to come in the
            line **before** :func:`attrs.define` or :func:`dataclasses.dataclass`.
        """


class Attrs:
    """
    Handler for "attrs" classes.
    """

    @staticmethod
    def check(cls: type) -> bool:
        try:
            import attrs

            return attrs.has(cls)
        except ImportError:
            return False

    @staticmethod
    def iter_fields(cls: type) -> types.OptionList:
        import attrs

        result: List[types.OptionInfo] = []

        def iter_attribs(r_cls: type, prefix: str) -> None:
            # Resolve types, optionally using the globals we stored in the class in
            # ".cls_attrs.combine()":
            r_cls = attrs.resolve_types(
                r_cls, globalns=getattr(r_cls, "__globals__", None)
            )
            for field in attrs.fields(r_cls):  # type: ignore[misc]
                if field.init is False:
                    continue
                if field.type is not None and attrs.has(field.type):
                    iter_attribs(field.type, f"{prefix}{field.name}.")
                else:
                    is_nothing = field.default is attrs.NOTHING
                    is_factory = isinstance(field.default, cast(type, attrs.Factory))
                    metadata = _get_metadata(field.metadata.get(constants.METADATA_KEY))
                    oinfo = types.OptionInfo(
                        parent_cls=r_cls,
                        path=f"{prefix}{field.name}",
                        cls=field.type,
                        is_secret=(
                            isinstance(field.repr, types.SecretRepr)
                            or (
                                isinstance(field.type, type)
                                and issubclass(field.type, types.SECRETS_TYPES)
                            )
                        ),
                        default=field.default,
                        has_no_default=is_nothing,
                        default_is_factory=is_factory,
                        converter=field.converter,
                        metadata=metadata,
                    )
                    result.append(oinfo)

        iter_attribs(cls, "")
        return tuple(result)

    @staticmethod
    def fields_to_parent_classes(cls: type) -> Dict[str, type]:
        import attrs

        return {
            field.name: (field.type if attrs.has(field.type) else cls)
            for field in attrs.fields(cls)  # type: ignore[misc]
        }

    @staticmethod
    def asdict(inst: Any) -> types.SettingsDict:
        import attrs

        return attrs.asdict(inst)

    @staticmethod
    def resolve_types(
        cls: Type[types.T],
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        include_extras: bool = True,
    ) -> Type[types.T]:
        import attrs

        return attrs.resolve_types(  # type: ignore[type-var]
            cls, globalns=globalns, localns=localns, include_extras=include_extras
        )


class Dataclasses:
    """
    Handler for :mod:`dataclasses` classes.
    """

    @staticmethod
    def check(cls: type) -> bool:
        return dataclasses.is_dataclass(cls)

    @classmethod
    def iter_fields(self, cls: type) -> types.OptionList:
        result: List[types.OptionInfo] = []

        def iter_attribs(r_cls: type, prefix: str) -> None:
            r_cls = self.resolve_types(r_cls)  # type: ignore[type-var]
            for field in dataclasses.fields(r_cls):
                if field.init is False:
                    continue
                if field.type is not None and dataclasses.is_dataclass(field.type):
                    iter_attribs(field.type, f"{prefix}{field.name}.")
                else:
                    is_nothing = field.default is dataclasses.MISSING
                    is_factory = (
                        is_nothing and field.default_factory is not dataclasses.MISSING
                    )
                    metadata = _get_metadata(field.metadata.get(constants.METADATA_KEY))
                    oinfo = types.OptionInfo(
                        parent_cls=r_cls,
                        path=f"{prefix}{field.name}",
                        cls=field.type,
                        is_secret=(
                            isinstance(field.repr, types.SecretRepr)
                            or (
                                isinstance(field.type, type)
                                and issubclass(field.type, types.SECRETS_TYPES)
                            )
                        ),
                        default=field.default,
                        has_no_default=is_nothing,
                        default_is_factory=is_factory,
                        converter=None,
                        metadata=metadata,
                    )
                    result.append(oinfo)

        iter_attribs(cls, "")
        return tuple(result)

    @staticmethod
    def fields_to_parent_classes(cls: type) -> Dict[str, type]:
        return {
            field.name: (field.type if dataclasses.is_dataclass(field.type) else cls)
            for field in dataclasses.fields(cls)
        }

    @staticmethod
    def asdict(inst: Any) -> types.SettingsDict:
        return dataclasses.asdict(inst)

    @staticmethod
    def resolve_types(
        cls: Type[types.T],
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        include_extras: bool = True,
    ) -> Type[types.T]:
        # Since calling get_type_hints is expensive we cache whether we've
        # done it already.
        if getattr(cls, "__dataclass_types_resolved__", None) != cls:
            import typing

            kwargs: Dict[str, Any] = {"globalns": globalns, "localns": localns}

            if PY_39:
                kwargs["include_extras"] = include_extras

            hints = typing.get_type_hints(cls, **kwargs)
            for field in dataclasses.fields(cls):  # type: ignore[arg-type]
                if field.name in hints:  # pragma: no cover
                    # Since fields have been frozen we must work around it.
                    object.__setattr__(field, "type", hints[field.name])
            # We store the class we resolved so that subclasses know they haven't
            # been resolved.
            cls.__dataclass_types_resolved__ = cls  # type: ignore[attr-defined]

        # Return the class so you can use it as a decorator too.
        return cls


class Pydantic:
    """
    Handler for "Pydantic" classes.
    """

    @staticmethod
    def check(cls: type) -> bool:
        try:
            import pydantic

            return issubclass(cls, pydantic.BaseModel)
        except ImportError:
            return False

    @staticmethod
    def iter_fields(cls: type) -> types.OptionList:
        import pydantic

        result: List[types.OptionInfo] = []

        def iter_attribs(r_cls: type, prefix: str) -> None:
            for name, field in r_cls.model_fields.items():  # type: ignore[attr-defined]
                if (
                    field.annotation is not None
                    and isinstance(field.annotation, type)
                    and issubclass(field.annotation, pydantic.BaseModel)
                ):
                    iter_attribs(field.annotation, f"{prefix}{name}.")
                else:
                    json_schema_extra = field.json_schema_extra or {}
                    metadata_or_none = json_schema_extra.get(constants.METADATA_KEY, {})
                    metadata = _get_metadata(metadata_or_none, field.description)

                    oinfo = types.OptionInfo(
                        parent_cls=r_cls,
                        path=f"{prefix}{name}",
                        cls=field.annotation,  # type: ignore[arg-type]
                        is_secret=(
                            isinstance(field.annotation, type)
                            and (
                                issubclass(
                                    field.annotation,
                                    (
                                        pydantic.SecretBytes,
                                        pydantic.SecretStr,
                                        *types.SECRETS_TYPES,
                                    ),
                                )
                            )
                        ),
                        default=field.default,
                        has_no_default=field.is_required(),
                        default_is_factory=False,
                        converter=None,
                        metadata=metadata,
                    )
                    result.append(oinfo)

        iter_attribs(cls, "")
        return tuple(result)

    @staticmethod
    def fields_to_parent_classes(cls: type) -> Dict[str, type]:
        import pydantic

        return {
            name: (
                field.annotation
                if isinstance(field.annotation, type)
                and issubclass(field.annotation, pydantic.BaseModel)
                else cls
            )
            for name, field in cls.model_fields.items()  # type: ignore[attr-defined]
        }

    @staticmethod
    def asdict(inst: Any) -> types.SettingsDict:
        return inst.dict()

    @staticmethod
    def resolve_types(
        cls: Type[types.T],
        globalns: Optional[Dict[str, Any]] = None,
        localns: Optional[Dict[str, Any]] = None,
        include_extras: bool = True,
    ) -> Type[types.T]:
        # Pydantic classes automatically resolve themselves.
        return cls


CLASS_HANDLERS: List[Type[ClsHandler]] = [
    Attrs,
    Dataclasses,
    Pydantic,
]


def handler_exists(cls: type) -> bool:
    """
    Check if a class handler for *cls* exist.

    Args:
        cls: The settings class to check the existence of a handler for.

    Return:
        ``True`` if there is a handler, otherwise ``False``.
    """
    for cls_handler in CLASS_HANDLERS:
        if cls_handler.check(cls):
            return True

    return False


def find_handler(cls: type) -> Type[ClsHandler]:
    """
    Return the proper class handler for *cls*.

    Args:
        cls: The settings class to find a handler for.

    Return:
        A :class:`ClsHandler` that works with *cls*.

    Raise:
        TypeError: If no class handler can be found for *cls*.
    """
    for cls_handler in CLASS_HANDLERS:
        if cls_handler.check(cls):
            return cls_handler

    raise TypeError(f"Cannot handle type: {cls}")


def deep_options(cls: type) -> types.OptionList:
    """
    Recursively iterates *cls* and nested attrs classes and returns a flat
    list of *(path, Attribute, type)* tuples.

    Args:
        cls: The class whose attributes will be listed.

    Returns:
        The flat list of attributes of *cls* and possibly nested attrs classes.
        *path* is a dot (``.``) separted path to the attribute, e.g.
        ``"parent_attr.child_attr.grand_child_attr``.

    Raises:
        NameError: if the type annotations can not be resolved.  This is, e.g., the
            case when recursive classes are being used.
    """
    cls_handler = find_handler(cls)
    return cls_handler.iter_fields(cls)


def group_options(
    cls: type, options: types.OptionList
) -> List[Tuple[type, types.OptionList]]:
    """
    Group (nested) options by parent class.

    If *cls* does not contain nested settings classes, return a single group for *cls*
    with all its options.

    If *cls* only contains nested subclasses, return one group per class containing all
    of that classes (posibly nested) options.

    If *cls* has multiple attributtes with the same nested settings class, create one
    group per attribute.

    If *cls* contains a mix of scalar options and nested options, return a mix of both.
    Scalar options schould be grouped (on top or bottom) or else multiple groups for the
    main settings class will be created.

    See the tests for details.

    Args:
        cls: The settings class
        options: The list of all options of the settings class.

    Return:
        A list of tuples matching a grouper class to all settings within that group.
    """
    cls_handler = find_handler(cls)
    fields_to_parents = cls_handler.fields_to_parent_classes(cls)

    def keyfn(o: types.OptionInfo) -> Tuple[str, type]:
        """
        Group by prefix and also return the corresponding group class.
        """
        basename, *remainder = o.path.split(".")
        prefix = basename if remainder else ""
        return prefix, fields_to_parents[basename]

    grouper = groupby(options, key=keyfn)
    grouped_options = [(g_cls[1], tuple(g_opts)) for g_cls, g_opts in grouper]
    return grouped_options


@overload
def resolve_types(
    cls: None = None,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    include_extras: bool = True,
) -> Callable[[Type[types.T]], Type[types.T]]: ...


@overload
def resolve_types(
    cls: Type[types.T],
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    include_extras: bool = True,
) -> Type[types.T]: ...


def resolve_types(
    cls: Optional[Type[types.T]] = None,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    include_extras: bool = True,
) -> Union[Type[types.T], Callable[[Type[types.T]], Type[types.T]]]:
    """
    Resolve any strings and forward annotations in type annotations.

    This is only required if you need concrete types in fields' *type* field. In other
    words, you don't need to resolve your types if you only use them for static type
    checking.

    With no arguments, names will be looked up in the module in which the class was
    created. If this is not what you want, e.g. if the name only exists inside a method,
    you may pass *globalns* or *localns* to specify other dictionaries in which to look
    up these names. See the docs of `typing.get_type_hints` for more details.

    Args:
        cls: Class to resolve.
        globalns: Dictionary containing global variables.
        localns: Dictionary containing local variables.
        include_extras: Resolve more accurately, if possible.
            Pass ``include_extras`` to ``typing.get_hints``, if supported by the typing
            module. On supported Python versions (3.9+), this resolves the types more
            accurately.

    Return:
        *cls* so you can use this function also as a class decorator.  Please note that
        you have to apply it **after** `attrs.define`. That means the decorator has to
        come in the line **before** `attrs.define`.

    Examples:
        ::

            >>> import typed_settings as ts
            >>>
            >>> @ts.settings
            ... class A:
            ...     opt: "int"
            ...
            >>> A = ts.resolve_types(A)
            >>>
            >>> @ts.resolve_types
            ... @ts.settings
            ... class B:
            ...     opt: "int"
            ...
            >>> @ts.resolve_types(globalns=globals(), localns=locals())
            ... @ts.settings
            ... class C:
            ...     opt: "int"
            ...

    .. versionadded:: 24.4.0
    """
    if cls is None:
        return functools.partial(  # type: ignore[return-value]
            resolve_types,
            globalns=globalns,
            localns=localns,
            include_extras=include_extras,
        )

    cls_handler = find_handler(cls)
    return cls_handler.resolve_types(
        cls, globalns=globalns, localns=localns, include_extras=include_extras
    )


def _get_metadata(metadata_or_none: Any, default_help: Optional[str] = None) -> dict:
    metadata = metadata_or_none if isinstance(metadata_or_none, dict) else {}

    cli_defaults: dict[str, Any] = {}
    if default_help:
        cli_defaults["help"] = default_help
    if "help" in metadata:
        cli_defaults["help"] = metadata["help"]

    click_config = {
        **cli_defaults,
        **metadata.get(constants.CLICK_METADATA_KEY, {}),
    }
    argparse_config = {
        **cli_defaults,
        **metadata.get(constants.ARGPARSE_METADATA_KEY, {}),
    }
    if click_config:
        metadata[constants.CLICK_METADATA_KEY] = click_config
    if argparse_config:
        metadata[constants.ARGPARSE_METADATA_KEY] = argparse_config

    return metadata



"""
Deprecated alias for :mod:`typed_settings.argparse_utils`.
"""

import warnings

from .cli_argparse import *  # noqa


warnings.warn(
    "This module has been renamed to 'typed_settings.argparse_utils'.",
    DeprecationWarning,
    stacklevel=2,
)



"""
Utility functions for working settings dicts and serilizing nested settings.
"""

from typing import (
    Any,
    Generator,
    Sequence,
    Tuple,
    get_args,
)

from .cls_utils import deep_options, handler_exists
from .types import (
    LoadedSettings,
    LoadedValue,
    MergedSettings,
    OptionList,
    SettingsDict,
)


__all__ = [
    "iter_settings",
    "get_path",
    "set_path",
    "merge_settings",
    "update_settings",
    "flat2nested",
]


def is_mutable_sequence(val: Any) -> bool:
    return (
        hasattr(val, "__iter__")
        and hasattr(val, "__getitem__")
        and hasattr(val, "__setitem__")
    )


def iter_settings(
    dct: SettingsDict, options: OptionList
) -> Generator[Tuple[str, Any], None, None]:
    """
    Iterate over the (possibly nested) options dict *dct* and yield
    *(option_path, value)* tuples.

    Args:
        dct: The dict of settings as returned by a loader.
        options: The list of all available options for a settings class.

    Return:
        A generator yield *(opton_path, value)* tuples.
    """
    for option in options:
        try:
            option_value = get_path(dct, option.path)

            if is_mutable_sequence(option_value) and not isinstance(
                option_value, (str, bytes)
            ):
                # only sub iterate in if declaration and actual value are lists
                args = get_args(option.cls)

                if args != () and handler_exists(args[0]):
                    # Recurse if "list[NestedSettings]" is detected and if
                    # NestedSettings is, e.g., an attrs class.
                    sub_options = deep_options(args[0])

                    for idx, sub_dct in enumerate(option_value):
                        for path, value in iter_settings(sub_dct, sub_options):
                            yield f"{option.path}.{idx}.{path}", value
                else:
                    # list of scalars
                    for idx, value in enumerate(option_value):
                        yield f"{option.path}.{idx}", value
            else:
                yield option.path, option_value
        except (KeyError, IndexError):
            continue


def get_path(dct: SettingsDict, path: str) -> Any:
    """
    Performs a nested dict lookup for *path* and returns the result.

    Calling ``get_path(dct, "a.b")`` is equivalent to ``dict["a"]["b"]``.
    If a part of the path is a non-negative integer, it is treated as list index.
    Calling ``get_path(dct, "a.0.b")`` is therefore equivalent to ``dct["a"][0]["b"]``.

    Args:
        dct: The source dict
        path: The path to look up.  It consists of the dot-separated nested
          keys.

    Returns:
        The looked up value.

    Raises:
        KeyError: if a key in *path* does not exist.
        IndexError: if a index in *path* is out of range.
    """
    for part in path.split("."):
        if part.isnumeric():
            dct = dct[int(part)]  # type: ignore[index]
        else:
            dct = dct[part]
    return dct


def set_path(dct: SettingsDict, path: str, val: Any) -> None:
    """
    Sets a value to a nested dict and automatically creates missing dicts
    should they not exist.

    Calling ``set_path(dct, "a.b", 3)`` is equivalent to ``dict["a"]["b"]
    = 3``.
    If a part of the path is a non-negative integer, it is treated as list index.
    Calling ``set_path(dct, "a.0.b", 3)`` is therefore equivalent to
    ``dct["a"][0]["b"] = 3``.

    Args:
        dct: The dict that should contain the value
        path: The (nested) path, a dot-separated concatenation of keys.
        val: The value to set

    Raises:
        IndexError: if a index in *path* is out of range.
    """
    *parts, key = path.split(".")
    for part in parts:
        if part.isnumeric():
            dct = dct[int(part)]  # type: ignore[index]
        else:
            dct = dct.setdefault(part, {})

    if key.isnumeric():
        key = int(key)  # type: ignore[assignment]
    dct[key] = val


def merge_settings(
    options: OptionList, settings: Sequence[LoadedSettings]
) -> MergedSettings:
    """
    Merge a sequence of settings dicts to a flat dict that maps option paths to the
    corresponding option values.

    Args:
        options: The list of all available options.
        settings: A sequence of loaded settings.

    Return:
        A dict that maps option paths to :class:`.LoadedValue` instances.

    The simplified input settings look like this::

        [
            ("loader a", {"spam": 1, "eggs": True}),
            ("loader b", {"spam": 2, "nested": {"x": "test"}}),
        ]

    The simpliefied output looks like this::

        {
            "spam": ("loader b", 2),
            "eggs": ("loader a", True),
            "nested.x": ("loader b", "test"),
        }
    """
    rsettings = settings[::-1]
    merged_settings: MergedSettings = {}
    for option_info in options:
        for loaded_settings in rsettings:
            try:
                value = get_path(loaded_settings.settings, option_info.path)
            except KeyError:
                pass
            else:
                merged_settings[option_info.path] = LoadedValue(
                    value, loaded_settings.meta
                )
                break
    return merged_settings


def update_settings(
    merged_settings: MergedSettings, settings: SettingsDict
) -> MergedSettings:
    """
    Return a copy of *merged_settings* updated with the values from *settings*.

    The loader meta data is not changed.

    Args:
        merged_settings: The merged settnigs dict to be updated.
        settings: The settings dict with additional values.

    Return:
        A copy of the input merged settings updated with the values from *settings*.
    """
    updated: MergedSettings = {}
    for path, (value, meta) in merged_settings.items():
        try:
            value = get_path(settings, path)
        except KeyError:
            pass
        updated[path] = LoadedValue(value, meta)
    return updated


def flat2nested(merged_settings: MergedSettings) -> SettingsDict:
    """
    Convert the flat *merged_settings* to a nested settings dict.
    """
    settings: SettingsDict = {}
    for path, loaded_value in merged_settings.items():
        set_path(settings, path, loaded_value.value)
    return settings



"""
Tests for "typed_settings.cli_utils".
"""

from collections.abc import (
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
)
from typing import (
    Any,
    Dict,
    List,
    NewType,
    Optional,
    Tuple,
    Union,
    cast,
    get_args,
    get_origin,
)

import attrs
import pytest

from typed_settings import cli_utils, default_converter, types
from typed_settings._compat import PY_39, PY_310


NewInt = NewType("NewInt", int)


def test_nodefaulttype_singleton() -> None:
    """
    `NoDefaultType()`  is a singleton.
    """
    assert cli_utils.NoDefaultType() is cli_utils.NO_DEFAULT


def test_nodefaulttype_repr() -> None:
    """
    `NoDefaultType()` has a nice repr.
    """
    assert repr(cli_utils.NoDefaultType()) == "NO_DEFAULT"


def _none_or_default(default: Any, is_optional: bool) -> Any:
    """
    Return "None" for "NO_DEFAULT" for optional options.
    """
    if is_optional and (default is None or default is cli_utils.NO_DEFAULT):
        return None
    return default


def handle_int(type: type, default: Any, is_optional: bool) -> cli_utils.StrDict:
    """
    An example handler function for the test TypeHandler.
    """
    return {
        "type": type,
        "default": default,
        "is_optional": is_optional,
        "called": "special",
    }


class TypeHandler:
    """
    An example TypeHandler for testing.
    """

    def get_scalar_handlers(self) -> Dict[type, cli_utils.TypeHandlerFunc]:
        return {
            int: handle_int,
        }

    def handle_scalar(
        self,
        type: Optional[type],
        default: Any,
        is_optional: bool,
    ) -> cli_utils.StrDict:
        return {
            "type": type,
            "default": default,
            "is_optional": is_optional,
            "called": "scalar",
        }

    def handle_tuple(
        self,
        type_args_maker: cli_utils.TypeArgsMaker,
        args: Tuple[Any, ...],
        default: Optional[Tuple],
        is_optional: bool,
    ) -> cli_utils.StrDict:
        return {
            "type_args_maker": type_args_maker,
            "args": args,
            "default": default,
            "is_optional": is_optional,
            "called": "tuple",
        }

    def handle_collection(
        self,
        type_args_maker: cli_utils.TypeArgsMaker,
        args: Tuple[Any, ...],
        default: Optional[List[Any]],
        is_optional: bool,
    ) -> cli_utils.StrDict:
        return {
            "type_args_maker": type_args_maker,
            "args": args,
            "default": default,
            "is_optional": is_optional,
            "called": "collection",
        }

    def handle_mapping(
        self,
        type_args_maker: cli_utils.TypeArgsMaker,
        args: Tuple[Any, ...],
        default: Any,
        is_optional: bool,
    ) -> cli_utils.StrDict:
        return {
            "type_args_maker": type_args_maker,
            "args": args,
            "default": default,
            "is_optional": is_optional,
            "called": "mapping",
        }


class TestTypeArgsMaker:
    """
    Tests for "TypeArgsMaker".
    """

    @pytest.fixture
    def tam(self) -> cli_utils.TypeArgsMaker:
        return cli_utils.TypeArgsMaker(TypeHandler())

    @pytest.mark.parametrize("default", [3, None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_special(
        self,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.get_scalar_handlers()", then the correct
        handler and returns its results.
        """
        t = Optional[int] if is_optional else int
        result = tam.get_kwargs(t, default)
        assert result == {
            "type": int,
            "default": _none_or_default(default, is_optional),
            "is_optional": is_optional,
            "called": "special",
        }

    @pytest.mark.parametrize("default", [None, 2, NewInt(1), cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_newtype(
        self,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        "NewType" and "Optional[NewType]" can be used as option types.
        """
        t = Optional[NewInt] if is_optional else NewInt
        result = tam.get_kwargs(t, default)
        assert result == {
            "type": int,
            "default": _none_or_default(default, is_optional),
            "is_optional": is_optional,
            "called": "special",
        }

    @pytest.mark.parametrize("default", ["x", None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_scalar(
        self,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_scalar()" and returns its results.
        """
        t = Optional[str] if is_optional else str
        result = tam.get_kwargs(t, default)
        assert result == {
            "type": str,
            "default": _none_or_default(default, is_optional),
            "is_optional": is_optional,
            "called": "scalar",
        }

    @pytest.mark.parametrize("default", [(1, "x"), None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_tuple(
        self,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_tuple()" and returns its results.
        """
        t = Optional[Tuple[int, str]] if is_optional else Tuple[int, str]
        result = tam.get_kwargs(t, default)
        assert result == {
            "type_args_maker": tam,
            "args": (int, str),
            "default": _none_or_default(default, is_optional=True),
            "is_optional": is_optional,
            "called": "tuple",
        }

    def test_tuple_wrong_deault_len(
        self,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM raises an error if a tuple default has the wrong length.
        """
        with pytest.raises(TypeError, match="Default value must be of len 2: 3"):
            tam.get_kwargs(Tuple[int, str], (1, "x", True))

    @pytest.mark.skipif(not PY_39, reason="Needs Python 3.8")
    @pytest.mark.parametrize("default", [[1, 2], None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_listtuple(
        self,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_collection()" for list-like tuples
        and returns its results.
        """
        t = Optional[tuple[int, ...]] if is_optional else tuple[int, ...]
        result = tam.get_kwargs(t, default)
        assert result == {
            "type_args_maker": tam,
            "args": (int, ...),
            "default": _none_or_default(default, is_optional=True),
            "is_optional": is_optional,
            "called": "collection",
        }

    @pytest.mark.skipif(not PY_39, reason="Needs Python 3.8")
    @pytest.mark.parametrize(
        "ctype", [list, Sequence, MutableSequence, set, frozenset, MutableSet]
    )
    @pytest.mark.parametrize("default", [[1, 2], None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_collection(
        self,
        ctype: Any,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_collection()" and returns its
        results.
        """
        t = Optional[ctype[int]] if is_optional else ctype[int]
        result = tam.get_kwargs(t, default)
        assert result == {
            "type_args_maker": tam,
            "args": (int,),
            "default": _none_or_default(default, is_optional=True),
            "is_optional": is_optional,
            "called": "collection",
        }

    @pytest.mark.skipif(not PY_39, reason="Needs Python 3.8")
    @pytest.mark.parametrize("ctype", [dict, Mapping, MutableMapping])
    @pytest.mark.parametrize("default", [{"a": 1}, None, cli_utils.NO_DEFAULT])
    @pytest.mark.parametrize("is_optional", [True, False])
    def test_mapping(
        self,
        ctype: Any,
        default: Any,
        is_optional: bool,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_mapping()" and returns its results.
        """
        t = Optional[ctype[str, int]] if is_optional else ctype[str, int]
        result = tam.get_kwargs(t, default)
        assert result == {
            "type_args_maker": tam,
            "args": (str, int),
            "default": _none_or_default(default, is_optional),
            "is_optional": is_optional,
            "called": "mapping",
        }

    @pytest.mark.parametrize("default", ["x", None, cli_utils.NO_DEFAULT])
    def test_none(
        self,
        default: Any,
        tam: cli_utils.TypeArgsMaker,
    ) -> None:
        """
        TAM calls calls "TypeHandler.handle_scalar()" and returns its results.
        """
        result = tam.get_kwargs(None, default)
        assert result == {
            "type": None,
            "default": default,
            "is_optional": False,
            "called": "scalar",
        }

    def test_unsupported(self, tam: cli_utils.TypeArgsMaker) -> None:
        """
        TAM raises a TypeError if it encounters an unsupported type.
        """
        with pytest.raises(TypeError, match="Cannot create CLI option for"):
            tam.get_kwargs(Union[int, str], 3)


@pytest.mark.parametrize(
    "name, cls, default, settings, expected",
    [
        (
            "a",
            int,
            attrs.NOTHING,
            {"a": types.LoadedValue(3, types.LoaderMeta("Dummy"))},
            3,
        ),
        ("a", int, attrs.NOTHING, {}, cli_utils.NO_DEFAULT),
        ("a", int, 2, {}, 2),
        ("a", List[int], attrs.Factory(list), {}, None),
        (
            "a",
            None,
            attrs.NOTHING,
            {"a": types.LoadedValue("3", types.LoaderMeta("Dummy"))},
            "3",
        ),
    ],
)
def test_get_default(
    name: str,
    cls: type,
    default: object,
    settings: dict,
    expected: object,
) -> None:
    """
    "get_default()" returns the loaded setting if possible or else the field's
    default value.
    """
    converter = default_converter()
    oinfo = types.OptionInfo(
        parent_cls=type,
        path=name,
        cls=cls,
        default=default,
        has_no_default=default is attrs.NOTHING,
        default_is_factory=isinstance(default, cast(type, attrs.Factory)),
    )
    result = cli_utils.get_default(oinfo, settings, converter)
    if isinstance(default, attrs.Factory):  # type: ignore[arg-type]
        assert isinstance(result, cli_utils.DefaultFactorySentinel)
        assert result() is expected
    else:
        assert result == expected


def test_get_default_factory() -> None:
    """
    Default factories are not invoked to generate a default value.
    """

    def factory(self: None) -> str:
        pytest.fail("This should not be invoked")

    default = attrs.Factory(factory, takes_self=True)
    attrs.Attribute(  # type: ignore[call-arg,var-annotated]
        "test", default, None, None, None, None, None, None, type=str
    )
    oinfo = types.OptionInfo(
        parent_cls=type,
        path="a",
        cls=str,
        default=default,
        has_no_default=False,
        default_is_factory=True,
    )
    result = cli_utils.get_default(oinfo, {}, default_converter())
    assert isinstance(result, cli_utils.DefaultFactorySentinel)
    assert result() is None


def test_get_default_cattrs_error() -> None:
    """
    "get_default()" checks if cattrs can convert a loaded default.
    """
    converter = default_converter()
    attrs.Attribute(  # type: ignore[call-arg,var-annotated]
        "test",
        attrs.NOTHING,
        None,
        None,
        None,
        None,
        None,
        None,
        type=List[int],
    )
    oinfo = types.OptionInfo(
        parent_cls=type,
        path="test",
        cls=List[int],
        default=attrs.NOTHING,
        has_no_default=True,
        default_is_factory=False,
    )
    with pytest.raises(ValueError, match="Invalid default .* for option 'test'"):
        cli_utils.get_default(
            oinfo,
            {"test": types.LoadedValue(["spam"], types.LoaderMeta("Dummy"))},
            converter,
        )


OPTIONAL_TEST_DATA = [
    (int, 3, (int, 3, None, (), False)),
    (int, cli_utils.NO_DEFAULT, (int, cli_utils.NO_DEFAULT, None, (), False)),
    (Optional[int], 3, (int, 3, None, (), True)),
    (Optional[int], None, (int, None, None, (), True)),
    (Optional[int], cli_utils.NO_DEFAULT, (int, None, None, (), True)),
    (Union[int, None], 3, (int, 3, None, (), True)),
    (Union[int, None], None, (int, None, None, (), True)),
    (Union[int, None], cli_utils.NO_DEFAULT, (int, None, None, (), True)),
    (Union[None, int], 3, (int, 3, None, (), True)),
    (Union[None, int], None, (int, None, None, (), True)),
    (Union[None, int], cli_utils.NO_DEFAULT, (int, None, None, (), True)),
    (List[int], None, (List[int], None, list, (int,), False)),
    (Optional[List[int]], None, (List[int], None, list, (int,), True)),
    (None, None, (None, None, None, (), False)),
]
if PY_310:
    OPTIONAL_TEST_DATA += [  # type: ignore[misc]  # type: ignore[misc]
        (int | None, 3, (int, 3, None, (), True)),
        (int | None, None, (int, None, None, (), True)),
        (int | None, cli_utils.NO_DEFAULT, (int, None, None, (), True)),
    ]


@pytest.mark.parametrize("t, d, expected", OPTIONAL_TEST_DATA)
def test_is_optional(
    t: Optional[type],
    d: Any,
    expected: Tuple[Optional[type], Any, Any, Tuple[Any, ...], bool],
) -> None:
    """
    Check if optional detects "Optional[T]", Union[T, None], and "T | None".
    """
    result = cli_utils.check_if_optional(t, d, get_origin(t), get_args(t))
    assert result == expected



"""
Helpers for invoking the 1Password CLI.
"""

import json
import subprocess
from typing import Any, Dict, Optional


__all__ = [
    "run",
    "get_item",
    "get_resource",
]


def run(*args: str) -> str:
    """
    Run ``op`` with the given arguments and return its stdout.

    Args:
        args: The command line arguments to pass to ``op``.

    Return:
        The stripped *stdout* of the ``op`` invocation.

    Raise:
        ValueError: If the CLI is not properly installed or the invocation
            failed.
    """
    cmd = ("op", *args)
    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)  # noqa: S603
    except FileNotFoundError:
        raise ValueError(
            "The 1Password CLI is not properly installed.  You can find help "
            "here: https://developer.1password.com/docs/cli/"
        ) from None
    except subprocess.CalledProcessError as e:
        _level, _date, _time, msg = e.stderr.strip().split(" ", maxsplit=3)
        raise ValueError(f'"op" error: {msg}') from None
    return result.stdout.strip()


def get_item(item: str, vault: Optional[str] = None) -> Dict[str, Any]:
    """
    Get *item* from 1Password.

    Args:
        item: The name of the item to retrieve.
        vault: Restrict search to this vault.

    Return:
        A dict mapping item labels to their values.

    Raise:
        ValueError: If the CLI is not properly installed or the invocation
            failed.
    """
    cmd = ["item", "get", "--format=json", item]
    if vault:
        cmd.append(f"--vault={vault}")
    data = json.loads(run(*cmd))
    return {
        field["label"]: field["value"] for field in data["fields"] if "value" in field
    }


def get_resource(resource: str) -> str:
    """
    Get the specified resource from 1Password using ``op read``.

    Args:
        resource: The URL (prefixed with ``op://`` for the resource.

    Return:
        The resource's value.

    Raise:
        ValueError: If the CLI is not properly installed or the invocation
            failed.
    """
    value = run("read", resource)
    return value



"""
Tests for "typed_settings.types".
"""

from pathlib import Path
from typing import Any

import pytest
from typing_extensions import assert_type

from typed_settings import types


def test_auto_singleton() -> None:
    """
    `_Auto()`  is a singleton.
    """
    assert types._Auto() is types.AUTO


def test_auto_repr() -> None:
    """
    `_Auto()` has a nice repr.
    """
    assert repr(types._Auto()) == "AUTO"


@pytest.mark.parametrize("name", [type, "type"])
def test_loader_meta_str(name: Any) -> None:
    """
    "LoaderMeta" has a nice str repr (even though it is not an attrs cls ;-)).
    """
    lm = types.LoaderMeta(name, Path("spam"))
    assert str(lm) == "LoaderMeta('type', PosixPath('spam'))"


class TestSecretStr:
    """
    Tests for "SecretStr".
    """

    def test_repr(self) -> None:
        """
        The repr of a non empty string is seven "*".
        """
        secret = types.SecretStr("spam")
        assert repr(secret) == "'*******'"

    def test_empty_repr(self) -> None:
        """
        The repr of a non empty string is seven "*".
        """
        secret = types.SecretStr("")
        assert repr(secret) == "''"

    def test_str(self) -> None:
        """
        The str repr is the original string.
        """
        secret = types.SecretStr("spam")
        assert str(secret) == "spam"

    def test_print(self, capsys: pytest.CaptureFixture) -> None:
        """
        Printing reveals the value.
        """
        secret = types.SecretStr("spam")
        print(secret)
        assert capsys.readouterr().out == "spam\n"

    def test_locals(self, capsys: pytest.CaptureFixture) -> None:
        """
        Printing locals does not leak the secrets.
        """
        secret = types.SecretStr("spam")
        print(locals())
        assert "'secret': '*******'" in capsys.readouterr().out


class TestSecret:
    """
    Tests for "Secret".
    """

    def test_repr(self) -> None:
        """
        The repr of a non empty string is seven *.
        """
        secret = types.Secret("spam")
        assert_type(secret, types.Secret[str])
        assert repr(secret) == "Secret('*******')"

    def test_empty_repr(self) -> None:
        """
        The repr of a non empty string is seven "*".
        """
        secret = types.Secret("")
        assert repr(secret) == "Secret('')"

    @pytest.mark.parametrize("v", ["spam", 0, 3, [1], True, False])
    def test_str(self, v: Any) -> None:
        """
        The str repr is the original string.
        """
        secret = types.Secret(v)
        assert str(secret) == "*******"

    @pytest.mark.parametrize("v, t", [("", str), ([], list), ((), tuple)])
    def test_empty_str(self, v: Any, t: type) -> None:
        """
        The str repr of an empty collection secret is an empty string.
        """
        secret = types.Secret(v)
        assert str(secret) == str(v)

    def test_print(self, capsys: pytest.CaptureFixture) -> None:
        """
        Printing does not reveal the value.
        """
        secret = types.Secret("spam")
        print(secret)
        assert capsys.readouterr().out == "*******\n"

    def test_locals(self, capsys: pytest.CaptureFixture) -> None:
        """
        Printing locals does not leak the secret.
        """
        secret = types.Secret("spam")
        print(locals())
        assert "'secret': Secret('*******')" in capsys.readouterr().out

    @pytest.mark.parametrize("v", ["spam", 3, [1], True])
    def test_bool_true(self, v: Any) -> None:
        """
        The secret's bool representation is that of the wrapped secret.
        """
        assert bool(types.Secret(v))

    @pytest.mark.parametrize("v", ["", 0, [], False])
    def test_bool_false(self, v: Any) -> None:
        """
        The secret's bool representation is that of the wrapped secret.
        """
        assert not bool(types.Secret(v))

    def test_get_value(self) -> None:
        """
        The secret value must explicitly be retrieved.
        """
        secret = types.Secret(42)
        assert secret.get_secret_value() == 42



"""
Core functionality for loading settings.
"""

import logging
import os
from contextlib import contextmanager
from pathlib import Path
from types import MappingProxyType
from typing import (
    Any,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
)

from . import cls_utils, dict_utils
from .converters import Converter, default_converter
from .exceptions import InvalidSettingsError
from .loaders import EnvLoader, FileLoader, Loader, TomlFormat, _DefaultsLoader
from .processors import Processor
from .types import (
    AUTO,
    ST,
    LoadedSettings,
    LoaderMeta,
    MergedSettings,
    OptionDict,
    OptionInfo,
    OptionList,
    SettingsDict,
    _Auto,
)


__all__ = [
    "LOGGER",
    "SettingsState",
    "default_loaders",
    "load",
    "load_settings",
    "convert",
]


LOGGER = logging.getLogger("typed-settings")


class SettingsState(Generic[ST]):
    """
    A representation of Typed Settings' internal state and configuration.
    """

    def __init__(
        self,
        settings_cls: Type[ST],
        loaders: Sequence[Loader],
        processors: Sequence[Processor],
        converter: Converter,
        base_dir: Path,
    ) -> None:
        self._cls = settings_cls
        self._options = tuple(cls_utils.deep_options(settings_cls))
        self._options_by_name = MappingProxyType({o.path: o for o in self._options})
        self._loaders = loaders
        self._processors = processors
        self._converter = converter
        self._base_dir = base_dir

    @property
    def settings_class(self) -> Type[ST]:
        """
        The user's settings class.
        """
        return self._cls

    @property
    def options(self) -> OptionList:
        """
        All options the settings class (and nested sub classes) define.
        """
        return self._options

    @property
    def options_by_path(self) -> OptionDict:
        """
        All options the settings class (and nested sub classes) define, mapped by
        their dotted path (e.g., `nested_cls.option_name`).
        """
        return self._options_by_name

    @property
    def loaders(self) -> List[Loader]:
        """
        A copy of the list of all configured settings loaders.
        """
        return list(self._loaders)

    @property
    def processors(self) -> List[Processor]:
        """
        A copy of the list of all configured post processors.
        """
        return list(self._processors)

    @property
    def converter(self) -> Converter:
        """
        The configured converter.
        """
        return self._converter

    @property
    def cwd(self) -> Path:
        """
        The current working directory.
        """
        return self._base_dir


def default_loaders(
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    *,
    config_file_section: Union[str, _Auto] = AUTO,
    config_files_var: Union[None, str, _Auto] = AUTO,
    env_prefix: Union[None, str, _Auto] = AUTO,
    env_nested_delimiter: str = "_",
) -> List[Loader]:
    """
    Return a list of default settings loaders that are used by :func:`load()`.

    These loaders are:

    #. A :class:`.FileLoader` loader configured with the :class:`.TomlFormat`
    #. An :class:`.EnvLoader`

    The :class:`.FileLoader` will load files from *config_files* and from the
    environment variable *config_files_var*.

    Args:
        appname: Your application's name -- used to derive defaults for the
          remaining args.

        config_files: Load settings from these files.  The last one has the
          highest precedence.

        config_file_section: Name of your app's section in the config file.
          By default, use *appname* (in lower case and with "_" replaced by
          "-".

        config_files_var: Load list of settings files from this environment
          variable.  By default, use :code:`{APPNAME}_SETTINGS`.  Multiple
          paths have to be separated by ":".  The last file has the highest
          precedence.  All files listed in this var have higher precedence than
          files from *config_files*.

          Set to ``None`` to disable this feature.

        env_prefix: Load settings from environment variables with this prefix.
          By default, use *APPNAME_*.

          Set to ``None`` to disable loading env vars.

        env_nested_delimiter: Delimiter for concatenating attribute names of nested
            classes in env. var. names.

    Return:
        A list of :class:`.Loader` instances.
    """
    loaders: List[Loader] = []

    section = (
        appname.lower().replace("_", "-")
        if isinstance(config_file_section, _Auto)
        else config_file_section
    )
    var_name = (
        f"{appname.upper()}_SETTINGS".replace("-", "_")
        if isinstance(config_files_var, _Auto)
        else config_files_var
    )
    loaders.append(
        FileLoader(
            files=config_files,
            env_var=var_name,
            formats={"*.toml": TomlFormat(section)},
        )
    )

    if env_prefix is None:
        LOGGER.debug("Loading settings from env vars is disabled.")
    else:
        prefix = (
            f"{appname.upper()}_".replace("-", "_")
            if isinstance(env_prefix, _Auto)
            else env_prefix
        )
        loaders.append(EnvLoader(prefix=prefix, nested_delimiter=env_nested_delimiter))

    return loaders


def load(
    cls: Type[ST],
    appname: str,
    config_files: Iterable[Union[str, Path]] = (),
    *,
    config_file_section: Union[str, _Auto] = AUTO,
    config_files_var: Union[None, str, _Auto] = AUTO,
    env_prefix: Union[None, str, _Auto] = AUTO,
    env_nested_delimiter: str = "_",
    base_dir: Path = Path(),
) -> ST:
    """
    Load settings for *appname* and return an instance of *cls*.

    This function is a shortcut for :func:`load_settings()` with
    :func:`default_loaders()`.

    Settings are loaded from *config_files* and from the files specified
    via the *config_files_var* environment variable.  Settings can also be
    overridden via environment variables named like the corresponding setting
    and prefixed with *env_prefix*.

    Settings precedence (from lowest to highest priority):

    - Default value from *cls*
    - First file from *config_files*
    - ...
    - Last file from *config_files*
    - First file from *config_files_var*
    - ...
    - Last file from *config_files_var*
    - Environment variable :code:`{env_prefix}_{SETTING}`

    Config files (both, explicitly specified, and loaded from an environment
    variable) are optional by default.  You can prepend an ``!`` to their path
    to mark them as mandatory (e.g., `!/etc/credentials.toml`).  An error is
    raised if a mandatory file does not exist.

    Args:
        cls: Attrs class with default settings.

        appname: Your application's name.  Used to derive defaults for the
          remaining args.

        config_files: Load settings from these files.

        config_file_section: Name of your app's section in the config file.
          By default, use *appname* (in lower case and with "_" replaced by
          "-".

        config_files_var: Load list of settings files from this environment
          variable.  By default, use :code:`{APPNAME}_SETTINGS`.  Multiple
          paths have to be separated by ":".  The last file has the highest
          precedence.  All files listed in this var have higher precedence than
          files from *config_files*.

          Set to ``None`` to disable this feature.

        env_prefix: Load settings from environment variables with this prefix.
          By default, use *APPNAME_*.

          Set to ``None`` to disable loading env vars.

        env_nested_delimiter: Delimiter for concatenating attribute names of nested
            classes in env. var. names.

        base_dir: Base directory for resolving relative paths in default option values.

    Return:
        An instance of *cls* populated with settings from settings files and
        environment variables.

    Raise:
        UnknownFormatError: When no :class:`~typed_settings.loaders.FileFormat`
            is configured for a loaded file.
        ConfigFileNotFoundError: If *path* does not exist.
        ConfigFileLoadError: If *path* cannot be read/loaded/decoded.
        InvalidOptionsError: If invalid settings have been found.
        InvalidValueError: If a value cannot be converted to the correct type.

    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    """
    loaders = default_loaders(
        appname=appname,
        config_files=config_files,
        config_file_section=config_file_section,
        config_files_var=config_files_var,
        env_prefix=env_prefix,
        env_nested_delimiter=env_nested_delimiter,
    )
    converter = default_converter()
    state = SettingsState(cls, loaders, [], converter, base_dir)
    settings = _load_settings(state)
    return convert(settings, state)


def load_settings(
    cls: Type[ST],
    loaders: Sequence[Loader],
    *,
    processors: Sequence[Processor] = (),
    converter: Optional[Converter] = None,
    base_dir: Path = Path(),
) -> ST:
    """
    Load settings defined by the class *cls* and return an instance of it.

    Args:
        cls: Attrs class with options (and default values).
        loaders: A list of settings :class:`.Loader`'s.
        processors: A list of settings :class:`.Processor`'s.
        converter: An optional :class:`.Converter` used for converting option values to
            the required type.

            By default, :func:`.default_converter()` is used.
        base_dir: Base directory for resolving relative paths in default option values.

    Return:
        An instance of *cls* populated with settings from the defined loaders.

    Raise:
        TsError: Depending on the configured loaders, any subclass of this
            exception.

    .. versionchanged:: 23.0.0
       Made *converter* a keyword-only argument
    .. versionchanged:: 23.0.0
       Added the *processors* argument
    .. versionchanged:: 23.1.0
       Added the *base_dir* argument
    """
    if converter is None:
        converter = default_converter()
    state = SettingsState(cls, loaders, processors, converter, base_dir=base_dir)
    settings = _load_settings(state)
    return convert(settings, state)


def _load_settings(state: SettingsState) -> MergedSettings:
    """
    Loads settings for *options* and returns them as dict.

    This function makes it easier to extend settings since it returns a dict
    that can easily be updated.
    """
    loaders = [_DefaultsLoader(state.cwd), *state.loaders]
    loaded_settings: List[LoadedSettings] = []
    for loader in loaders:
        result = loader(state.settings_class, state.options)
        if isinstance(result, LoadedSettings):
            loaded_settings.append(result)
        else:
            loaded_settings.extend(result)

    merged_settings = dict_utils.merge_settings(state.options, loaded_settings)

    # Get a "dict view" to merged settings and update the merged_settings afterwards
    # without changing the LoaderMeta for each setting
    settings_dict = dict_utils.flat2nested(merged_settings)
    for processor in state.processors:
        settings_dict = processor(settings_dict, state.settings_class, state.options)
    merged_settings = dict_utils.update_settings(merged_settings, settings_dict)

    return merged_settings


def convert(
    merged_settings: MergedSettings,
    state: SettingsState[ST],
) -> ST:
    """
    Create an instance of *cls* from the settings in *merged_settings*.

    Args:
        merged_settings: The loaded and merged settings by settings name.
        state: The state and configuration for this run.

    Return:
        An instance of *cls*.

    Raise:
        InvalidSettingsError: If an instance of *cls* cannot be created for the given
            settings.
    """
    settings_dict: SettingsDict = {}
    errors: List[str] = []
    loaded_settings_paths: Set[str] = set()
    oi_by_path = state.options_by_path
    for path, (value, meta) in merged_settings.items():
        oinfo = oi_by_path[path]
        try:
            converted_value = convert_value(oinfo, value, meta, state.converter)
        except Exception as e:
            errors.append(
                f"Could not convert value {value!r} for option "
                f"{path!r} from loader {meta.name}: {e!r}"
            )
            continue
        dict_utils.set_path(settings_dict, path, converted_value)
        loaded_settings_paths.add(path)

    for option_info in state.options:
        if option_info.path in loaded_settings_paths:
            continue
        if option_info.has_default:
            continue
        errors.append(f"No value set for required option {option_info.path!r}")

    try:
        settings = state.converter.structure(settings_dict, state.settings_class)
    except Exception as e:
        errors.append(f"Could not convert loaded settings: {e!r}")

    if errors:
        errs = "".join(f"\n- {e}" for e in errors)
        raise InvalidSettingsError(
            f"{len(errors)} errors occured while converting the loaded option values "
            f"to an instance of {state.settings_class.__name__!r}:{errs}"
        )

    return settings


def convert_value(
    oinfo: OptionInfo, value: Any, meta: LoaderMeta, converter: Converter
) -> Any:
    """
    Convert the value for an option to the designated type.

    Args:
        oinfo: Metadata for the option.
        value: The value to be converted.
        meta: Metadata for the loader that loaded *value*.
        converter: The converter to use.

    Return:
        The converted value.

    Raise:
        Exception: If the value cannot be converted.
    """
    if oinfo.cls:
        with _set_context(meta):
            if oinfo.converter:
                converted_value = oinfo.converter(value)
            else:
                converted_value = converter.structure(value, oinfo.cls)
    else:
        converted_value = value
    return converted_value


@contextmanager
def _set_context(meta: LoaderMeta) -> Generator[None, None, None]:
    """
    Set the context for converting option values from a given loader.

    Currently only chagnes the cwd to :attr:`.LoaderMeta.cwd`.

    Args:
        meta: A loaders meta data

    Return:
        A context manager (that yields ``None``)
    """
    old_cwd = os.getcwd()
    os.chdir(meta.base_dir)
    try:
        yield
    finally:
        os.chdir(old_cwd)



"""
Test that all public functions are properly exposed.
"""

from pathlib import Path
from typing import Type

import pytest

import typed_settings as ts


@ts.settings
class Settings:
    """
    A simple test class with the TS aliases for attrs.
    """

    u: str = ts.option()
    p: str = ts.secret()


@ts.settings(frozen=True)
class FrozenSettings:
    """
    A frozen test class with the TS aliases for attrs.
    """

    u: str = ts.option()
    p: str = ts.secret()


classes = [Settings, FrozenSettings]


@pytest.mark.parametrize("cls", classes)
def test_load(cls: Type[Settings], tmp_path: Path) -> None:
    """
    We can load settings with a class decorated with our decorator.
    """
    f = tmp_path.joinpath("cfg.toml")
    f.write_text('[test]\nu = "spam"\np = "eggs"\n')
    settings = ts.load(cls, "test", [f])
    assert settings == cls("spam", "eggs")


@pytest.mark.parametrize("cls", classes)
def test_load_settings(cls: Type[Settings], tmp_path: Path) -> None:
    """
    We can load settings with a class decorated with our decorator.
    """
    f = tmp_path.joinpath("cfg.toml")
    f.write_text('[test]\nu = "spam"\np = "eggs"\n')
    settings = ts.load(cls, "test", [f])
    assert settings == cls("spam", "eggs")


def test_dir() -> None:
    """
    dir(typed_settings) returns the expected list of names, including the
    ones requiring optional dependencies.
    """
    names = dir(ts)
    assert names == [
        "EnvLoader",
        "FileLoader",
        "Secret",
        "SecretStr",
        "SettingsState",
        "TomlFormat",
        "cli",
        "click_options",
        "combine",
        "convert",
        "default_converter",
        "default_loaders",
        "evolve",
        "find",
        "load",
        "load_settings",
        "option",
        "pass_settings",
        "register_strlist_hook",
        "resolve_types",
        "secret",
        "settings",
    ]
